@@ -5,6 +5,7 @@
 #include <gio/gio.h>
 
 #include "manifest.h"
+#include "nbd.h"
 #include "utils.h"
 
 #define R_BUNDLE_ERROR r_bundle_error_quark()
@@ -17,14 +18,30 @@ typedef enum {
 	R_BUNDLE_ERROR_UNSAFE,
 	R_BUNDLE_ERROR_PAYLOAD,
 	R_BUNDLE_ERROR_FORMAT,
+	R_BUNDLE_ERROR_VERITY,
+	R_BUNDLE_ERROR_CRYPT,
 } RBundleError;
 
+typedef struct {
+	gchar *tls_cert;
+	gchar *tls_key;
+	gchar *tls_ca;
+	gboolean tls_no_verify;
+	GStrv http_headers;
+} RaucBundleAccessArgs;
+
 typedef struct {
 	gchar *path;
 	gchar *origpath;
 	gchar *storepath;
+
+	RaucNBDDevice *nbd_dev;
+	RaucNBDServer *nbd_srv;
+
 	GInputStream *stream;
+
 	goffset size;
+	GBytes *enveloped_data;
 	GBytes *sigdata;
 	gchar *mount_point;
 	RaucManifest *manifest;
@@ -32,6 +49,7 @@ typedef struct {
 	gboolean signature_verified;
 	gboolean payload_verified;
 	gboolean exclusive_verified;
+	gboolean was_encrypted;
 	gchar *exclusive_check_error;
 	STACK_OF(X509) *verified_chain;
 } RaucBundle;
@@ -71,16 +89,20 @@ G_GNUC_WARN_UNUSED_RESULT;
  *
  * This will verify and check the bundle content.
  *
+ * For bundle formats with detached CMS (i.e. 'verity' or 'crypt'),
+ * this will also initialize the manifest.
+ *
  * @param bundlename filename of the bundle to check
  * @param bundle return location for a RaucBundle struct.
  *               This will contain all bundle information obtained by
  *               check_bundle
  * @param params bit-field enum CheckBundleParams with additional flags for the check
+ * @param access_args Optional arguments to control streaming access
  * @param error Return location for a GError
  *
  * @return TRUE on success, FALSE if an error occurred
  */
-gboolean check_bundle(const gchar *bundlename, RaucBundle **bundle, CheckBundleParams params, GError **error)
+gboolean check_bundle(const gchar *bundlename, RaucBundle **bundle, CheckBundleParams params, RaucBundleAccessArgs *access_args, GError **error)
 G_GNUC_WARN_UNUSED_RESULT;
 
 /**
@@ -203,6 +225,18 @@ G_GNUC_WARN_UNUSED_RESULT;
 gboolean create_casync_bundle(RaucBundle *bundle, const gchar *outbundle, GError **error)
 G_GNUC_WARN_UNUSED_RESULT;
 
+/**
+ * Encrypt a crypt bundle.
+ *
+ * @param bundle RaucBundle struct as returned by check_bundle()
+ * @param outbundle output location for encrypted crypt bundle
+ * @param error Return location for a GError
+ *
+ * @return TRUE on success, FALSE if an error occurred
+ */
+gboolean encrypt_bundle(RaucBundle *bundle, const gchar *outbundle, GError **error)
+G_GNUC_WARN_UNUSED_RESULT;
+
 /**
  * Mount a bundle.
  *
@@ -238,3 +272,11 @@ gboolean umount_bundle(RaucBundle *bundle, GError **error);
 void free_bundle(RaucBundle *bundle);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC(RaucBundle, free_bundle);
+
+/**
+ * Frees the memory pointed to by the RaucBundleAccessArgs, but not the
+ * structure itself.
+ *
+ * @param access_args RaucBundleAccessArgs to clear
+ */
+void clear_bundle_access_args(RaucBundleAccessArgs *access_args);
