@@ -19,6 +19,7 @@ typedef enum {
 	R_CONFIG_ERROR_DUPLICATE_BOOTNAME,
 	R_CONFIG_ERROR_SLOT_TYPE,
 	R_CONFIG_ERROR_INVALID_DEVICE,
+	R_CONFIG_ERROR_DATA_DIRECTORY,
 } RConfigError;
 
 #define R_CONFIG_ERROR r_config_error_quark()
@@ -50,10 +51,14 @@ typedef struct {
 	gchar *mount_prefix;
 	gchar *store_path;
 	gchar *tmp_path;
+	gchar *casync_install_args;
+	gboolean use_desync;
 	gboolean activate_installed;
+	gchar *data_directory;
 	gchar *statusfile_path;
 	gchar *keyring_path;
 	gchar *keyring_directory;
+	gboolean keyring_allow_partial_chain;
 	gboolean keyring_check_crl;
 	gchar *keyring_check_purpose;
 	gboolean use_bundle_signing_time;
@@ -66,6 +71,16 @@ typedef struct {
 
 	gchar *systeminfo_handler;
 
+	/* streaming */
+	gchar *streaming_sandbox_user;
+	gchar *streaming_tls_cert;
+	gchar *streaming_tls_key;
+	gchar *streaming_tls_ca;
+
+	/* encryption */
+	gchar *encryption_key;
+	gchar *encryption_cert;
+
 	GHashTable *slots;
 } RaucConfig;
 
