@@ -41,8 +41,10 @@ typedef struct {
 	gchar *keyringpath;
 	gchar *keyringdirectory;
 	gchar *signing_keyringpath;
+	gchar *encryption_key;
 	gchar *mksquashfs_args;
 	gchar *casync_args;
+	gchar **recipients;
 	gchar **intermediatepaths;
 	/* optional global mount prefix overwrite */
 	gchar *mountprefix;
