@@ -5,6 +5,18 @@
 
 #define BIT(nr) (1UL << (nr))
 
+/* Evaluate EXPRESSION, and repeat as long as it returns -1 with `errno'
+ * set to EINTR. Needed for builds against musl, taken from glibc's unistd.h.
+ */
+#ifndef TEMP_FAILURE_RETRY
+#define TEMP_FAILURE_RETRY(expression) \
+	(__extension__ \
+		 ({ long int __result; \
+		    do {__result = (long int) (expression);} \
+		    while (__result == -1L && errno == EINTR); \
+		    __result; }))
+#endif
+
 /* Use
  *
  *   g_auto(filedesc) fd = -1
@@ -125,9 +137,24 @@ gboolean rm_tree(const gchar *path, GError **error);
  * @return An absolute path name, determined as described above, NULL if undeterminable
  *         [transfer full]
  */
-gchar *resolve_path(const gchar *basefile, gchar *path)
+gchar *resolve_path(const gchar *basefile, const gchar *path)
 G_GNUC_WARN_UNUSED_RESULT;
 
+/**
+ * Resolve path based on directory of `basefile` argument or current working dir
+ * and free path.
+ *
+ * This is a wrapper around resolve_path(), for use when the path argument is
+ * not needed after the call.
+ *
+ * @param basefile Reference path to resolve `path` to
+ * @param path The path to resolve an absolute path for (freed)
+ *
+ * @return An absolute path name, determined as described above, NULL if undeterminable
+ *         [transfer full]
+ */
+gchar *resolve_path_take(const gchar *basefile, gchar *path)
+G_GNUC_WARN_UNUSED_RESULT;
 
 gboolean check_remaining_groups(GKeyFile *key_file, GError **error)
 G_GNUC_WARN_UNUSED_RESULT;
@@ -179,3 +206,20 @@ guint8 *r_hex_decode(const gchar *hex, size_t len)
 G_GNUC_WARN_UNUSED_RESULT;
 gchar *r_hex_encode(const guint8 *raw, size_t len)
 G_GNUC_WARN_UNUSED_RESULT;
+
+gboolean r_read_exact(const int fd, guint8 *data, size_t size, GError **error)
+G_GNUC_WARN_UNUSED_RESULT;
+gboolean r_write_exact(const int fd, const guint8 *data, size_t size, GError **error)
+G_GNUC_WARN_UNUSED_RESULT;
+
+gboolean r_pread_exact(const int fd, guint8 *data, size_t size, off_t offset, GError **error)
+G_GNUC_WARN_UNUSED_RESULT;
+
+gboolean r_pwrite_exact(const int fd, const guint8 *data, size_t size, off_t offset, GError **error)
+G_GNUC_WARN_UNUSED_RESULT;
+
+gboolean r_pwrite_lazy(const int fd, const guint8 *data, size_t size, off_t offset, GError **error)
+G_GNUC_WARN_UNUSED_RESULT;
+
+guint get_sectorsize(gint fd)
+G_GNUC_WARN_UNUSED_RESULT;
