@@ -24,7 +24,8 @@ void default_config(RaucConfig **config)
 	 */
 	c->bundle_formats_mask =
 		1 << R_MANIFEST_FORMAT_PLAIN |
-		        1 << R_MANIFEST_FORMAT_VERITY;
+		        1 << R_MANIFEST_FORMAT_VERITY |
+		        1 << R_MANIFEST_FORMAT_CRYPT;
 
 	*config = c;
 }
@@ -103,6 +104,8 @@ gboolean parse_bundle_formats(guint *mask, const gchar *config, GError **error)
 			format = R_MANIFEST_FORMAT_PLAIN;
 		} else if (g_strcmp0(token, "verity") == 0) {
 			format = R_MANIFEST_FORMAT_VERITY;
+		} else if (g_strcmp0(token, "crypt") == 0) {
+			format = R_MANIFEST_FORMAT_CRYPT;
 		} else {
 			g_set_error(
 					error,
@@ -159,9 +162,9 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	g_autoptr(RaucConfig) c = g_new0(RaucConfig, 1);
 	gboolean res = FALSE;
 	g_autoptr(GKeyFile) key_file = NULL;
-	gchar **groups;
+	g_auto(GStrv) groups = NULL;
 	gsize group_count;
-	GList *slotlist = NULL;
+	g_autoptr(GList) slotlist = NULL;
 	GHashTable *slots = NULL;
 	g_autoptr(GHashTable) bootnames = NULL;
 	GList *l;
@@ -209,7 +212,7 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 		c->system_bb_statename = key_file_consume_string(key_file, "system", "barebox-statename", NULL);
 		c->system_bb_dtbpath = key_file_consume_string(key_file, "system", "barebox-dtbpath", NULL);
 	} else if (g_strcmp0(c->system_bootloader, "grub") == 0) {
-		c->grubenv_path = resolve_path(filename,
+		c->grubenv_path = resolve_path_take(filename,
 				key_file_consume_string(key_file, "system", "grubenv", NULL));
 		if (!c->grubenv_path) {
 			g_debug("No grubenv path provided, using /boot/grub/grubenv as default");
@@ -227,7 +230,7 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 		}
 		g_key_file_remove_key(key_file, "system", "efi-use-bootnext", NULL);
 	} else if (g_strcmp0(c->system_bootloader, "custom") == 0) {
-		c->custom_bootloader_backend = resolve_path(filename,
+		c->custom_bootloader_backend = resolve_path_take(filename,
 				key_file_consume_string(key_file, "handlers", "bootloader-custom-backend", NULL));
 		if (!c->custom_bootloader_backend) {
 			g_set_error(
@@ -301,7 +304,7 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	/* parse 'variant-file' key */
 	variant_data = key_file_consume_string(key_file, "system", "variant-file", &ierror);
 	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
-		variant_data = NULL;
+		g_clear_pointer(&variant_data, g_free);
 		g_clear_error(&ierror);
 	} else if (ierror) {
 		g_propagate_error(error, ierror);
@@ -348,10 +351,36 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 		c->system_variant = variant_data;
 	}
 
+	/* parse data/status location
+	 *
+	 * We have multiple levels of backwards compatibility:
+	 * - per-slot status and no shared data directory
+	 *   (by default or explicitly with ``statusfile=per-slot``)
+	 * - central status file and no shared data directory
+	 *   (``statusfile=/data/central.raucs``)
+	 * - central status file and shared data directory
+	 *   (``statusfile=/data/central.raucs`` and ``data-directory=/data/rauc``)
+	 * - central status file in shared data directory
+	 *   (``data-directory=/data/rauc``, implies ``statusfile=/data/rauc/central.rauc``)
+	 */
+	c->data_directory = resolve_path_take(filename,
+			key_file_consume_string(key_file, "system", "data-directory", &ierror));
+	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
+		g_clear_error(&ierror);
+	} else if (ierror) {
+		g_propagate_error(error, ierror);
+		res = FALSE;
+		goto free;
+	}
+
 	c->statusfile_path = key_file_consume_string(key_file, "system", "statusfile", &ierror);
 	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
-		g_message("Config option 'statusfile=<path>/per-slot' unset, falling back to per-slot status");
-		c->statusfile_path = g_strdup("per-slot");
+		if (c->data_directory) {
+			c->statusfile_path = g_build_filename(c->data_directory, "central.raucs", NULL);
+		} else {
+			g_message("Config option 'statusfile=<path>/per-slot' unset, falling back to per-slot status");
+			c->statusfile_path = g_strdup("per-slot");
+		}
 		g_clear_error(&ierror);
 	} else if (ierror) {
 		g_propagate_error(error, ierror);
@@ -360,6 +389,15 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	}
 
 	if (g_strcmp0(c->statusfile_path, "per-slot") == 0) {
+		if (c->data_directory) {
+			g_set_error(
+					error,
+					R_CONFIG_ERROR,
+					R_CONFIG_ERROR_DATA_DIRECTORY,
+					"Using data-directory= with statusfile=per-slot is not supported.");
+			res = FALSE;
+			goto free;
+		}
 		g_message("Using per-slot statusfile");
 	} else {
 		gchar *resolved = resolve_path(filename, c->statusfile_path);
@@ -371,7 +409,8 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	/* parse bundle formats */
 	c->bundle_formats_mask =
 		1 << R_MANIFEST_FORMAT_PLAIN |
-		        1 << R_MANIFEST_FORMAT_VERITY;
+		        1 << R_MANIFEST_FORMAT_VERITY |
+		        1 << R_MANIFEST_FORMAT_CRYPT;
 	bundle_formats = key_file_consume_string(key_file, "system", "bundle-formats", &ierror);
 	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
 		g_clear_error(&ierror);
@@ -395,9 +434,9 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	g_key_file_remove_group(key_file, "system", NULL);
 
 	/* parse [keyring] section */
-	c->keyring_path = resolve_path(filename,
+	c->keyring_path = resolve_path_take(filename,
 			key_file_consume_string(key_file, "keyring", "path", NULL));
-	c->keyring_directory = resolve_path(filename,
+	c->keyring_directory = resolve_path_take(filename,
 			key_file_consume_string(key_file, "keyring", "directory", NULL));
 
 	c->keyring_check_crl = g_key_file_get_boolean(key_file, "keyring", "check-crl", &ierror);
@@ -412,6 +451,18 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	}
 	g_key_file_remove_key(key_file, "keyring", "check-crl", NULL);
 
+	c->keyring_allow_partial_chain = g_key_file_get_boolean(key_file, "keyring", "allow-partial-chain", &ierror);
+	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND) ||
+	    g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_GROUP_NOT_FOUND)) {
+		c->keyring_allow_partial_chain = FALSE;
+		g_clear_error(&ierror);
+	} else if (ierror) {
+		g_propagate_error(error, ierror);
+		res = FALSE;
+		goto free;
+	}
+	g_key_file_remove_key(key_file, "keyring", "allow-partial-chain", NULL);
+
 	c->use_bundle_signing_time = g_key_file_get_boolean(key_file, "keyring", "use-bundle-signing-time", &ierror);
 	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND) ||
 	    g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_GROUP_NOT_FOUND)) {
@@ -445,6 +496,18 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	/* parse [casync] section */
 	c->store_path = key_file_consume_string(key_file, "casync", "storepath", NULL);
 	c->tmp_path = key_file_consume_string(key_file, "casync", "tmppath", NULL);
+	c->casync_install_args = key_file_consume_string(key_file, "casync", "install-args", NULL);
+	c->use_desync = g_key_file_get_boolean(key_file, "casync", "use-desync", &ierror);
+	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND) ||
+	    g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_GROUP_NOT_FOUND)) {
+		c->use_desync = FALSE;
+		g_clear_error(&ierror);
+	} else if (ierror) {
+		g_propagate_error(error, ierror);
+		res = FALSE;
+		goto free;
+	}
+	g_key_file_remove_key(key_file, "casync", "use-desync", NULL);
 	if (!check_remaining_keys(key_file, "casync", &ierror)) {
 		g_propagate_error(error, ierror);
 		res = FALSE;
@@ -452,8 +515,32 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	}
 	g_key_file_remove_group(key_file, "casync", NULL);
 
+	/* parse [streaming] section */
+	c->streaming_sandbox_user = key_file_consume_string(key_file, "streaming", "sandbox-user", NULL);
+	c->streaming_tls_cert = key_file_consume_string(key_file, "streaming", "tls-cert", NULL);
+	c->streaming_tls_key = key_file_consume_string(key_file, "streaming", "tls-key", NULL);
+	c->streaming_tls_ca = key_file_consume_string(key_file, "streaming", "tls-ca", NULL);
+	if (!check_remaining_keys(key_file, "streaming", &ierror)) {
+		g_propagate_error(error, ierror);
+		res = FALSE;
+		goto free;
+	}
+	g_key_file_remove_group(key_file, "streaming", NULL);
+
+	/* parse [encryption] section */
+	c->encryption_key = resolve_path_take(filename,
+			key_file_consume_string(key_file, "encryption", "key", NULL));
+	c->encryption_cert = resolve_path_take(filename,
+			key_file_consume_string(key_file, "encryption", "cert", NULL));
+	if (!check_remaining_keys(key_file, "encryption", &ierror)) {
+		g_propagate_error(error, ierror);
+		res = FALSE;
+		goto free;
+	}
+	g_key_file_remove_group(key_file, "encryption", NULL);
+
 	/* parse [autoinstall] section */
-	c->autoinstall_path = resolve_path(filename,
+	c->autoinstall_path = resolve_path_take(filename,
 			key_file_consume_string(key_file, "autoinstall", "path", NULL));
 	if (!check_remaining_keys(key_file, "autoinstall", &ierror)) {
 		g_propagate_error(error, ierror);
@@ -463,13 +550,13 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 	g_key_file_remove_group(key_file, "autoinstall", NULL);
 
 	/* parse [handlers] section */
-	c->systeminfo_handler = resolve_path(filename,
+	c->systeminfo_handler = resolve_path_take(filename,
 			key_file_consume_string(key_file, "handlers", "system-info", NULL));
 
-	c->preinstall_handler = resolve_path(filename,
+	c->preinstall_handler = resolve_path_take(filename,
 			key_file_consume_string(key_file, "handlers", "pre-install", NULL));
 
-	c->postinstall_handler = resolve_path(filename,
+	c->postinstall_handler = resolve_path_take(filename,
 			key_file_consume_string(key_file, "handlers", "post-install", NULL));
 	if (!check_remaining_keys(key_file, "handlers", &ierror)) {
 		g_propagate_error(error, ierror);
@@ -517,11 +604,16 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 			slot->name = g_intern_string(value);
 			g_free(value);
 
+			/* If we have a data_directory, use a slot.<class>.<index>
+			 * subdirectory for per-slot data. */
+			if (c->data_directory)
+				slot->data_directory = g_build_filename(c->data_directory, groups[i], NULL);
+
 			slot->description = key_file_consume_string(key_file, groups[i], "description", NULL);
 
 			slot->sclass = g_intern_string(groupsplit[1]);
 
-			value = resolve_path(filename,
+			value = resolve_path_take(filename,
 					key_file_consume_string(key_file, groups[i], "device", &ierror));
 			if (!value) {
 				g_propagate_error(error, ierror);
@@ -644,7 +736,8 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 			g_key_file_remove_key(key_file, groups[i], "resize", NULL);
 
 			if (g_strcmp0(slot->type, "boot-mbr-switch") == 0 ||
-			    g_strcmp0(slot->type, "boot-gpt-switch") == 0) {
+			    g_strcmp0(slot->type, "boot-gpt-switch") == 0 ||
+			    g_strcmp0(slot->type, "boot-raw-fallback") == 0) {
 				slot->region_start = key_file_consume_binary_suffixed_string(key_file, groups[i],
 						"region-start", &ierror);
 				if (ierror) {
@@ -713,7 +806,6 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 			goto free;
 		}
 	}
-	g_list_free(slotlist);
 
 	if (!fix_grandparent_links(slots, &ierror)) {
 		g_propagate_error(error, ierror);
@@ -729,8 +821,6 @@ gboolean load_config(const gchar *filename, RaucConfig **config, GError **error)
 		goto free;
 	}
 
-	g_strfreev(groups);
-
 	res = TRUE;
 free:
 	if (res)
@@ -762,7 +852,10 @@ void free_config(RaucConfig *config)
 		return;
 
 	g_free(config->system_compatible);
+	g_free(config->system_variant);
 	g_free(config->system_bootloader);
+	g_free(config->system_bb_statename);
+	g_free(config->system_bb_dtbpath);
 	g_free(config->mount_prefix);
 	g_free(config->store_path);
 	g_free(config->tmp_path);
@@ -775,6 +868,12 @@ void free_config(RaucConfig *config)
 	g_free(config->systeminfo_handler);
 	g_free(config->preinstall_handler);
 	g_free(config->postinstall_handler);
+	g_free(config->streaming_sandbox_user);
+	g_free(config->streaming_tls_cert);
+	g_free(config->streaming_tls_key);
+	g_free(config->streaming_tls_ca);
+	g_free(config->encryption_key);
+	g_free(config->encryption_cert);
 	g_clear_pointer(&config->slots, g_hash_table_destroy);
 	g_free(config);
 }
@@ -1022,13 +1121,14 @@ void load_slot_status(RaucSlot *dest_slot)
 {
 	g_return_if_fail(dest_slot);
 
-	if (dest_slot->status)
-		return;
+	if (!dest_slot->status) {
+		if (g_strcmp0(r_context()->config->statusfile_path, "per-slot") == 0)
+			load_slot_status_locally(dest_slot);
+		else
+			load_slot_status_globally();
+	}
 
-	if (g_strcmp0(r_context()->config->statusfile_path, "per-slot") == 0)
-		load_slot_status_locally(dest_slot);
-	else
-		load_slot_status_globally();
+	r_slot_clean_data_directory(dest_slot);
 }
 
 static gboolean save_slot_status_locally(RaucSlot *dest_slot, GError **error)
@@ -1111,6 +1211,8 @@ gboolean save_slot_status(RaucSlot *dest_slot, GError **error)
 	g_return_val_if_fail(dest_slot, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
+	r_slot_clean_data_directory(dest_slot);
+
 	if (g_strcmp0(r_context()->config->statusfile_path, "per-slot") == 0)
 		return save_slot_status_locally(dest_slot, error);
 	else
