@@ -1,3 +1,4 @@
+#include <errno.h>
 #include <gio/gio.h>
 #include <string.h>
 
@@ -271,6 +272,20 @@ gboolean r_context_configure(GError **error)
 			break;
 	}
 
+	if (context->config->data_directory) {
+		if (g_mkdir_with_parents(context->config->data_directory, 0700) != 0) {
+			int err = errno;
+			g_set_error(
+					error,
+					G_FILE_ERROR,
+					g_file_error_from_errno(err),
+					"Failed to create data directory '%s': %s",
+					context->config->data_directory,
+					g_strerror(err));
+			return FALSE;
+		}
+	}
+
 	if (context->config->system_variant_type == R_CONFIG_SYS_VARIANT_DTB) {
 		gchar *compatible = get_system_dtb_compatible(&ierror);
 		if (!compatible) {
@@ -296,7 +311,7 @@ gboolean r_context_configure(GError **error)
 		gchar *key = NULL;
 		gchar *value = NULL;
 
-		vars = g_hash_table_new(g_str_hash, g_str_equal);
+		vars = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
 
 		g_message("Getting Systeminfo: %s", context->config->systeminfo_handler);
 		res = launch_and_wait_variables_handler(context->config->systeminfo_handler, vars, &ierror);
@@ -339,6 +354,10 @@ gboolean r_context_configure(GError **error)
 		context->config->keyring_directory = g_strdup(context->keyringdirectory);
 	}
 
+	if (context->encryption_key) {
+		context->config->encryption_key = g_strdup(context->encryption_key);
+	}
+
 	context->pending = FALSE;
 
 	return TRUE;
@@ -392,6 +411,11 @@ static void r_context_send_progress(gboolean op_finished, gboolean success)
 		iter = g_list_next(iter);
 	}
 
+	/* This step is not 100% itself, so it must not be the root step even though the
+	   previous steps sum to 100. Max out the percentage at 99% in that case. */
+	if (step->percent_done < 100.0f && percentage > 99.0f)
+		percentage = 99.0f;
+
 	g_assert_cmpint(percentage, <=, 100);
 
 	/* call installer callback with percentage and message */
@@ -637,13 +661,16 @@ void r_context_install_info_free(RContextInstallationInfo *info)
 void r_context_clean(void)
 {
 	if (context) {
+		g_clear_pointer(&context->configpath, g_free);
 		g_clear_pointer(&context->certpath, g_free);
 		g_clear_pointer(&context->keypath, g_free);
 		g_clear_pointer(&context->keyringpath, g_free);
 		g_clear_pointer(&context->keyringdirectory, g_free);
 		g_clear_pointer(&context->signing_keyringpath, g_free);
+		g_clear_pointer(&context->encryption_key, g_free);
 		g_clear_pointer(&context->mksquashfs_args, g_free);
 		g_clear_pointer(&context->casync_args, g_free);
+		g_clear_pointer(&context->recipients, g_strfreev);
 		g_clear_pointer(&context->intermediatepaths, g_strfreev);
 		g_clear_pointer(&context->mountprefix, g_free);
 		g_clear_pointer(&context->bootslot, g_free);
@@ -652,8 +679,8 @@ void r_context_clean(void)
 
 		g_clear_pointer(&context->install_info, r_context_install_info_free);
 
-		if (context->config) {
-			context->config->keyring_path = NULL;
-		}
+		g_clear_pointer(&context->config, free_config);
+
+		g_clear_pointer(&context, g_free);
 	}
 }
