@@ -480,7 +480,7 @@ static gchar* parse_handler_output(gchar* line)
 	return message;
 }
 
-static gboolean verify_compatible(RaucInstallArgs *args, RaucManifest *manifest)
+static gboolean verify_compatible(RaucInstallArgs *args, RaucManifest *manifest, GError **error)
 {
 	if (args->ignore_compatible) {
 		return TRUE;
@@ -488,7 +488,8 @@ static gboolean verify_compatible(RaucInstallArgs *args, RaucManifest *manifest)
 			manifest->update_compatible) == 0) {
 		return TRUE;
 	} else {
-		g_warning("incompatible manifest for this system (%s): %s",
+		g_set_error(error, R_INSTALL_ERROR, R_INSTALL_ERROR_COMPAT_MISMATCH,
+				"Compatible mismatch: Expected '%s' but bundle manifest has '%s'",
 				r_context()->config->system_compatible,
 				manifest->update_compatible);
 		return FALSE;
@@ -721,6 +722,7 @@ out:
 
 static gboolean launch_and_wait_custom_handler(RaucInstallArgs *args, gchar* bundledir, RaucManifest *manifest, GHashTable *target_group, GError **error)
 {
+	GError *ierror = NULL;
 	g_autofree gchar* handler_name = NULL;
 	gboolean res = FALSE;
 
@@ -728,16 +730,14 @@ static gboolean launch_and_wait_custom_handler(RaucInstallArgs *args, gchar* bun
 
 	/* Allow overriding compatible check by hook */
 	if (manifest->hooks.install_check) {
-		GError *ierror = NULL;
 		run_bundle_hook(manifest, bundledir, "install-check", &ierror);
 		if (ierror) {
 			g_propagate_error(error, ierror);
 			res = FALSE;
 			goto out;
 		}
-	} else if (!verify_compatible(args, manifest)) {
-		g_set_error_literal(error, R_INSTALL_ERROR, R_INSTALL_ERROR_COMPAT_MISMATCH,
-				"Compatible mismatch");
+	} else if (!verify_compatible(args, manifest, &ierror)) {
+		g_propagate_error(error, ierror);
 		res = FALSE;
 		goto out;
 	}
@@ -849,10 +849,9 @@ static gboolean launch_and_wait_default_handler(RaucInstallArgs *args, gchar* bu
 			res = FALSE;
 			goto early_out;
 		}
-	} else if (!verify_compatible(args, manifest)) {
+	} else if (!verify_compatible(args, manifest, &ierror)) {
+		g_propagate_error(error, ierror);
 		res = FALSE;
-		g_set_error_literal(error, R_INSTALL_ERROR, R_INSTALL_ERROR_COMPAT_MISMATCH,
-				"Compatible mismatch");
 		goto early_out;
 	}
 
@@ -1073,12 +1072,18 @@ gboolean do_install_bundle(RaucInstallArgs *args, GError **error)
 	// TODO: mount info in context ?
 	install_args_update(args, "Checking and mounting bundle...");
 
-	res = check_bundle(bundlefile, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(bundlefile, &bundle, CHECK_BUNDLE_DEFAULT, &args->access_args, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
 		goto out;
 	}
 
+	if (bundle->manifest && bundle->manifest->bundle_format == R_MANIFEST_FORMAT_CRYPT && !bundle->was_encrypted) {
+		g_set_error(error, R_INSTALL_ERROR, R_INSTALL_ERROR_REJECTED, "Refusing to install unencrypted crypt bundles");
+		res = FALSE;
+		goto out;
+	}
+
 	res = mount_bundle(bundle, &ierror);
 	if (!res) {
 		g_propagate_prefixed_error(
@@ -1202,6 +1207,7 @@ void install_args_free(RaucInstallArgs *args)
 	g_mutex_clear(&args->status_mutex);
 	g_assert_cmpint(args->status_result, >=, 0);
 	g_assert_true(g_queue_is_empty(&args->status_messages));
+	clear_bundle_access_args(&args->access_args);
 	g_free(args);
 }
 
