@@ -32,14 +32,19 @@ gboolean trust_environment = FALSE;
 gboolean verification_disabled = FALSE;
 gboolean no_check_time = FALSE;
 gboolean info_dumpcert = FALSE;
+gboolean info_dumprecipients = FALSE;
 gboolean status_detailed = FALSE;
 gchar *output_format = NULL;
+gchar *keypath = NULL;
+gchar *certpath = NULL;
 gchar *signing_keyring = NULL;
 gchar *mksquashfs_args = NULL;
 gchar *casync_args = NULL;
+gchar **recipients = NULL;
 gchar *handler_args = NULL;
 gchar *bootslot = NULL;
 gboolean utf8_supported = FALSE;
+RaucBundleAccessArgs access_args = {0};
 
 static gchar* make_progress_line(gint percentage)
 {
@@ -236,12 +241,32 @@ static gboolean install_start(int argc, char **argv)
 	args->status_result = 2;
 
 	args->ignore_compatible = install_ignore_compatible;
+	if (access_args.tls_cert)
+		args->access_args.tls_cert = g_strdup(access_args.tls_cert);
+	if (access_args.tls_key)
+		args->access_args.tls_key = g_strdup(access_args.tls_key);
+	if (access_args.tls_ca)
+		args->access_args.tls_ca = g_strdup(access_args.tls_ca);
+	if (access_args.tls_no_verify)
+		args->access_args.tls_no_verify = access_args.tls_no_verify;
+	if (access_args.http_headers)
+		args->access_args.http_headers = g_strdupv(access_args.http_headers);
 
 	r_loop = g_main_loop_new(NULL, FALSE);
 	if (ENABLE_SERVICE) {
 		g_auto(GVariantDict) dict = G_VARIANT_DICT_INIT(NULL);
 
 		g_variant_dict_insert(&dict, "ignore-compatible", "b", args->ignore_compatible);
+		if (args->access_args.tls_cert)
+			g_variant_dict_insert(&dict, "tls-cert", "s", args->access_args.tls_cert);
+		if (args->access_args.tls_key)
+			g_variant_dict_insert(&dict, "tls-key", "s", args->access_args.tls_key);
+		if (args->access_args.tls_ca)
+			g_variant_dict_insert(&dict, "tls-ca", "s", args->access_args.tls_ca);
+		if (args->access_args.tls_no_verify)
+			g_variant_dict_insert(&dict, "tls-no-verify", "b", args->access_args.tls_no_verify);
+		if (args->access_args.http_headers)
+			g_variant_dict_insert(&dict, "http-headers", "^as", args->access_args.http_headers);
 
 		installer = r_installer_proxy_new_for_bus_sync(bus_type,
 				G_DBUS_PROXY_FLAGS_GET_INVALIDATED_PROPERTIES,
@@ -458,7 +483,7 @@ static gboolean write_slot_start(int argc, char **argv)
 		goto out;
 	}
 
-	g_message("Slot written successfully");
+	g_print("Slot written successfully\n");
 
 out:
 	return TRUE;
@@ -503,7 +528,7 @@ static gboolean resign_start(int argc, char **argv)
 	if (no_check_time)
 		check_bundle_params |= CHECK_BUNDLE_NO_CHECK_TIME;
 
-	if (!check_bundle(argv[2], &bundle, check_bundle_params, &ierror)) {
+	if (!check_bundle(argv[2], &bundle, check_bundle_params, NULL, &ierror)) {
 		g_printerr("%s\n", ierror->message);
 		g_clear_error(&ierror);
 		r_exit_status = 1;
@@ -521,6 +546,7 @@ out:
 	return TRUE;
 }
 
+G_GNUC_UNUSED
 static gboolean replace_signature_start(int argc, char **argv)
 {
 	CheckBundleParams check_bundle_params = CHECK_BUNDLE_DEFAULT;
@@ -561,7 +587,7 @@ static gboolean replace_signature_start(int argc, char **argv)
 	if (trust_environment)
 		check_bundle_params |= CHECK_BUNDLE_TRUST_ENV;
 
-	if (!check_bundle(argv[2], &bundle, check_bundle_params, &ierror)) {
+	if (!check_bundle(argv[2], &bundle, check_bundle_params, NULL, &ierror)) {
 		g_printerr("%s\n", ierror->message);
 		g_clear_error(&ierror);
 		r_exit_status = 1;
@@ -579,8 +605,10 @@ out:
 	return TRUE;
 }
 
+G_GNUC_UNUSED
 static gboolean extract_signature_start(int argc, char **argv)
 {
+	CheckBundleParams check_bundle_params = CHECK_BUNDLE_DEFAULT;
 	RaucBundle *bundle = NULL;
 	GError *ierror = NULL;
 	g_debug("extract signature start");
@@ -606,7 +634,10 @@ static gboolean extract_signature_start(int argc, char **argv)
 	g_debug("input bundle: %s", argv[2]);
 	g_debug("output file: %s", argv[3]);
 
-	if (!check_bundle(argv[2], &bundle, CHECK_BUNDLE_DEFAULT, &ierror)) {
+	if (trust_environment)
+		check_bundle_params |= CHECK_BUNDLE_TRUST_ENV;
+
+	if (!check_bundle(argv[2], &bundle, check_bundle_params, NULL, &ierror)) {
 		g_printerr("%s\n", ierror->message);
 		g_clear_error(&ierror);
 		r_exit_status = 1;
@@ -626,6 +657,7 @@ out:
 
 static gboolean extract_start(int argc, char **argv)
 {
+	CheckBundleParams check_bundle_params = CHECK_BUNDLE_DEFAULT;
 	RaucBundle *bundle = NULL;
 	GError *ierror = NULL;
 	g_debug("extract start");
@@ -651,7 +683,10 @@ static gboolean extract_start(int argc, char **argv)
 	g_debug("input bundle: %s", argv[2]);
 	g_debug("output dir: %s", argv[3]);
 
-	if (!check_bundle(argv[2], &bundle, CHECK_BUNDLE_DEFAULT, &ierror)) {
+	if (trust_environment)
+		check_bundle_params |= CHECK_BUNDLE_TRUST_ENV;
+
+	if (!check_bundle(argv[2], &bundle, check_bundle_params, NULL, &ierror)) {
 		g_printerr("%s\n", ierror->message);
 		g_clear_error(&ierror);
 		r_exit_status = 1;
@@ -710,7 +745,7 @@ static gboolean convert_start(int argc, char **argv)
 	if (trust_environment)
 		check_bundle_params |= CHECK_BUNDLE_TRUST_ENV;
 
-	if (!check_bundle(argv[2], &bundle, check_bundle_params, &ierror)) {
+	if (!check_bundle(argv[2], &bundle, check_bundle_params, NULL, &ierror)) {
 		g_printerr("%s\n", ierror->message);
 		g_clear_error(&ierror);
 		r_exit_status = 1;
@@ -730,6 +765,60 @@ out:
 	return TRUE;
 }
 
+G_GNUC_UNUSED
+static gboolean encrypt_start(int argc, char **argv)
+{
+	RaucBundle *bundle = NULL;
+	GError *ierror = NULL;
+	g_debug("encrypt start");
+
+	if (r_context()->recipients == NULL) {
+		g_printerr("One or multiple recipient certificates must be provided (via --to)\n");
+		r_exit_status = 1;
+		goto out;
+	}
+
+	if (argc < 3) {
+		g_printerr("An input bundle must be provided\n");
+		r_exit_status = 1;
+		goto out;
+	}
+
+	if (argc < 4) {
+		g_printerr("An output bundle name must be provided\n");
+		r_exit_status = 1;
+		goto out;
+	}
+
+	if (argc > 4) {
+		g_printerr("Excess argument: %s\n", argv[4]);
+		r_exit_status = 1;
+		goto out;
+	}
+
+	g_debug("input bundle: %s", argv[2]);
+	g_debug("output bundle: %s", argv[3]);
+
+	if (!check_bundle(argv[2], &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror)) {
+		g_printerr("%s\n", ierror->message);
+		g_clear_error(&ierror);
+		r_exit_status = 1;
+		goto out;
+	}
+
+	if (!encrypt_bundle(bundle, argv[3], &ierror)) {
+		g_printerr("Failed to create bundle: %s\n", ierror->message);
+		g_clear_error(&ierror);
+		r_exit_status = 1;
+		goto out;
+	}
+
+	g_print("Encrypted bundle written to %s\n", argv[3]);
+
+out:
+	return TRUE;
+}
+
 /* Definition list for terminal colors */
 #define KNRM  "\x1B[0m"
 #define KRED  "\x1B[31m"
@@ -760,7 +849,7 @@ static gchar *info_formatter_shell(RaucManifest *manifest)
 {
 	GString *text = g_string_new(NULL);
 	GPtrArray *hooks = NULL;
-	gchar *hookstring = NULL;
+	gchar *temp_string = NULL;
 	gint cnt;
 
 	formatter_shell_append(text, "RAUC_MF_COMPATIBLE", manifest->update_compatible);
@@ -775,9 +864,9 @@ static gchar *info_formatter_shell(RaucManifest *manifest)
 	}
 	g_ptr_array_add(hooks, NULL);
 
-	hookstring = g_strjoinv(" ", (gchar**) hooks->pdata);
-	formatter_shell_append(text, "RAUC_MF_HOOKS", hookstring);
-	g_free(hookstring);
+	temp_string = g_strjoinv(" ", (gchar**) hooks->pdata);
+	formatter_shell_append(text, "RAUC_MF_HOOKS", temp_string);
+	g_free(temp_string);
 
 	g_ptr_array_unref(hooks);
 
@@ -802,11 +891,18 @@ static gchar *info_formatter_shell(RaucManifest *manifest)
 		}
 		g_ptr_array_add(hooks, NULL);
 
-		hookstring = g_strjoinv(" ", (gchar**) hooks->pdata);
-		formatter_shell_append_n(text, "RAUC_IMAGE_HOOKS", cnt, hookstring);
-		g_free(hookstring);
+		temp_string = g_strjoinv(" ", (gchar**) hooks->pdata);
+		formatter_shell_append_n(text, "RAUC_IMAGE_HOOKS", cnt, temp_string);
+		g_free(temp_string);
 
 		g_ptr_array_unref(hooks);
+
+		if (img->adaptive) {
+			temp_string = g_strjoinv(" ", (gchar**) img->adaptive);
+			formatter_shell_append_n(text, "RAUC_IMAGE_ADAPTIVE", cnt, temp_string);
+			g_free(temp_string);
+		}
+
 		cnt++;
 	}
 
@@ -817,7 +913,8 @@ static gchar *info_formatter_readable(RaucManifest *manifest)
 {
 	GString *text = g_string_new(NULL);
 	GPtrArray *hooks = NULL;
-	gchar *hookstring = NULL;
+	gchar *temp_string = NULL;
+	gboolean show_crypt_key = FALSE; /* change to TRUE to display dm-crypt key */
 	gint cnt;
 
 	g_string_append_printf(text, "Compatible: \t'%s'\n", manifest->update_compatible);
@@ -831,12 +928,23 @@ static gchar *info_formatter_readable(RaucManifest *manifest)
 	}
 	g_ptr_array_add(hooks, NULL);
 
-	hookstring = g_strjoinv(" ", (gchar**) hooks->pdata);
-	g_string_append_printf(text, "Hooks:      \t'%s'\n", hookstring);
-	g_free(hookstring);
+	temp_string = g_strjoinv(" ", (gchar**) hooks->pdata);
+	g_string_append_printf(text, "Hooks:      \t'%s'\n", temp_string);
+	g_free(temp_string);
 
-	g_string_append_printf(text, "Bundle Format: \t%s\n", r_manifest_bundle_format_to_str(manifest->bundle_format));
-	if (manifest->bundle_format == R_MANIFEST_FORMAT_VERITY) {
+	g_string_append_printf(text, "Bundle Format: \t%s", r_manifest_bundle_format_to_str(manifest->bundle_format));
+	if (manifest->bundle_format == R_MANIFEST_FORMAT_CRYPT) {
+		if (manifest->was_encrypted)
+			g_string_append_printf(text, KBLD KGRN " [encrypted CMS]"KNRM);
+		else
+			g_string_append_printf(text, KBLD KYEL " [unencrypted CMS]"KNRM);
+	}
+	g_string_append_printf(text, "\n");
+
+	if (manifest->bundle_format == R_MANIFEST_FORMAT_CRYPT) {
+		g_string_append_printf(text, "  Crypt Key: \t'%s'\n", show_crypt_key ? manifest->bundle_crypt_key : "<hidden>");
+	}
+	if (manifest->bundle_format == R_MANIFEST_FORMAT_VERITY || manifest->bundle_format == R_MANIFEST_FORMAT_CRYPT) {
 		g_string_append_printf(text, "  Verity Salt: \t'%s'\n", manifest->bundle_verity_salt);
 		g_string_append_printf(text, "  Verity Hash: \t'%s'\n", manifest->bundle_verity_hash);
 		g_string_append_printf(text, "  Verity Size: \t%"G_GUINT64_FORMAT "\n", manifest->bundle_verity_size);
@@ -872,12 +980,18 @@ static gchar *info_formatter_readable(RaucManifest *manifest)
 		}
 		g_ptr_array_add(hooks, NULL);
 
-		hookstring = g_strjoinv(" ", (gchar**) hooks->pdata);
-		g_string_append_printf(text, "\tHooks:     %s\n", hookstring);
-		g_free(hookstring);
+		temp_string = g_strjoinv(" ", (gchar**) hooks->pdata);
+		g_string_append_printf(text, "\tHooks:     %s\n", temp_string);
+		g_free(temp_string);
 
 		g_ptr_array_unref(hooks);
 
+		if (img->adaptive) {
+			temp_string = g_strjoinv(" ", (gchar**) img->adaptive);
+			g_string_append_printf(text, "\tAdaptive:  %s\n", temp_string);
+			g_free(temp_string);
+		}
+
 		cnt++;
 	}
 
@@ -942,6 +1056,14 @@ static gchar* info_formatter_json_base(RaucManifest *manifest, gboolean pretty)
 			json_builder_add_string_value(builder, "post-install");
 		}
 		json_builder_end_array(builder);
+		json_builder_set_member_name(builder, "adaptive");
+		json_builder_begin_array(builder);
+		if (img->adaptive) {
+			for (gchar **m = img->adaptive; *m != NULL; m++) {
+				json_builder_add_string_value(builder, *m);
+			}
+		}
+		json_builder_end_array(builder);
 		json_builder_end_object(builder);
 		json_builder_end_object(builder);
 	}
@@ -1017,7 +1139,7 @@ static gboolean info_start(int argc, char **argv)
 	if (no_check_time)
 		check_bundle_params |= CHECK_BUNDLE_NO_CHECK_TIME;
 
-	res = check_bundle(bundlelocation, &bundle, check_bundle_params, &error);
+	res = check_bundle(bundlelocation, &bundle, check_bundle_params, &access_args, &error);
 	if (!res) {
 		g_printerr("%s\n", error->message);
 		g_clear_error(&error);
@@ -1045,6 +1167,16 @@ static gboolean info_start(int argc, char **argv)
 		g_free(text);
 	}
 
+	if (info_dumprecipients) {
+		if (!bundle->enveloped_data) {
+			g_print("No recipient data to dump (bundle is not encrypted)\n\n");
+		} else {
+			text = envelopeddata_to_string(bundle->enveloped_data, NULL);
+			g_print("%s\n", text);
+			g_free(text);
+		}
+	}
+
 	if (!output_format || g_strcmp0(output_format, "readable") == 0) {
 		if (!bundle->verified_chain) {
 			g_print("Signature unverified\n");
@@ -1644,7 +1776,6 @@ static gboolean status_start(int argc, char **argv)
 	const gchar *slot_identifier = NULL;
 	GError *ierror = NULL;
 	gboolean res = FALSE;
-	RInstaller *proxy = NULL;
 	g_autoptr(RaucStatusPrint) status_print = NULL;
 
 	g_debug("status start");
@@ -1656,7 +1787,7 @@ static gboolean status_start(int argc, char **argv)
 			g_printerr("Failed to determine slot states: %s\n", ierror->message);
 			g_clear_error(&ierror);
 			r_exit_status = 1;
-			goto out;
+			return TRUE;
 		}
 
 		res = determine_boot_states(&ierror);
@@ -1678,7 +1809,7 @@ static gboolean status_start(int argc, char **argv)
 
 		status_print->primary = r_boot_get_primary(&ierror);
 		if (!status_print->primary) {
-			g_printerr("%s\n", ierror->message);
+			g_printerr("Failed getting primary slot: %s\n", ierror->message);
 			g_clear_error(&ierror);
 		}
 
@@ -1688,12 +1819,11 @@ static gboolean status_start(int argc, char **argv)
 		status_print->slots = r_context()->config->slots;
 	} else {
 		if (!retrieve_status_via_dbus(&status_print, &ierror)) {
-			message = g_strdup_printf(
-					"error retrieving slot status via D-Bus: %s",
+			g_printerr("Error retrieving slot status via D-Bus: %s\n",
 					ierror->message);
 			g_error_free(ierror);
 			r_exit_status = 1;
-			goto out;
+			return TRUE;
 		}
 	}
 
@@ -1701,15 +1831,15 @@ static gboolean status_start(int argc, char **argv)
 		if (!print_status(status_print)) {
 			r_exit_status = 1;
 		}
-		goto out;
+		return TRUE;
 	} else if (argc == 3) {
 		slot_identifier = "booted";
 	} else if (argc == 4) {
 		slot_identifier = argv[3];
 	} else { /* argc > 4 */
-		g_warning("Too many arguments");
+		g_printerr("Too many arguments\n");
 		r_exit_status = 1;
-		goto out;
+		return TRUE;
 	}
 
 	if (g_strcmp0(argv[2], "mark-good") == 0) {
@@ -1719,38 +1849,44 @@ static gboolean status_start(int argc, char **argv)
 	} else if (g_strcmp0(argv[2], "mark-active") == 0) {
 		state = "active";
 	} else {
-		g_message("unknown subcommand %s", argv[2]);
+		g_printerr("unknown subcommand %s\n", argv[2]);
 		r_exit_status = 1;
-		goto out;
+		return TRUE;
 	}
 
 	if (ENABLE_SERVICE) {
+		g_autoptr(RInstaller) proxy = NULL;
+
 		proxy = r_installer_proxy_new_for_bus_sync(bus_type,
 				G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
 				"de.pengutronix.rauc", "/", NULL, &ierror);
 		if (proxy == NULL) {
-			message = g_strdup_printf("rauc mark: error creating proxy: %s",
-					ierror->message);
+			if (g_dbus_error_is_remote_error(ierror))
+				g_dbus_error_strip_remote_error(ierror);
+			g_printerr("rauc mark: error creating proxy: %s\n", ierror->message);
 			g_error_free(ierror);
 			r_exit_status = 1;
-			goto out;
+			return TRUE;
 		}
 		g_debug("Trying to contact rauc service");
 		if (!r_installer_call_mark_sync(proxy, state, slot_identifier,
 				&slot_name, &message, NULL, &ierror)) {
-			message = g_strdup(ierror->message);
+			if (g_dbus_error_is_remote_error(ierror))
+				g_dbus_error_strip_remote_error(ierror);
+			g_printerr("rauc mark: %s\n", ierror->message);
 			g_error_free(ierror);
 			r_exit_status = 1;
-			goto out;
+			return TRUE;
 		}
 	} else {
 		r_exit_status = mark_run(state, slot_identifier, NULL, &message) ? 0 : 1;
+		if (r_exit_status)
+			g_printerr("rauc mark: %s\n", message);
+		return TRUE;
 	}
 
-out:
 	if (message)
-		g_message("rauc status: %s", message);
-	g_clear_object(&proxy);
+		g_print("rauc status: %s\n", message);
 
 	return TRUE;
 }
@@ -1787,7 +1923,7 @@ static gboolean mount_start(int argc, char **argv)
 		goto out; /* an error message was already printed by resolve_bundle_path */
 	g_debug("input bundle: %s", bundlelocation);
 
-	res = check_bundle(bundlelocation, &bundle, CHECK_BUNDLE_DEFAULT, &error);
+	res = check_bundle(bundlelocation, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &error);
 	if (!res) {
 		g_printerr("%s\n", error->message);
 		g_clear_error(&error);
@@ -1834,6 +1970,7 @@ typedef enum  {
 	EXTRACT_SIG,
 	EXTRACT,
 	CONVERT,
+	ENCRYPT,
 	STATUS,
 	INFO,
 	WRITE_SLOT,
@@ -1892,11 +2029,22 @@ static GOptionEntry entries_convert[] = {
 	{0}
 };
 
+static GOptionEntry entries_extract_signature[] = {
+	{"trust-environment", '\0', 0, G_OPTION_ARG_NONE, &trust_environment, "trust environment and skip bundle access checks", NULL},
+	{0}
+};
+
+static GOptionEntry entries_extract[] = {
+	{"trust-environment", '\0', 0, G_OPTION_ARG_NONE, &trust_environment, "trust environment and skip bundle access checks", NULL},
+	{0}
+};
+
 static GOptionEntry entries_info[] = {
 	{"no-verify", '\0', 0, G_OPTION_ARG_NONE, &verification_disabled, "disable bundle verification", NULL},
 	{"no-check-time", '\0', 0, G_OPTION_ARG_NONE, &no_check_time, "don't check validity period of certificates against current time", NULL},
 	{"output-format", '\0', 0, G_OPTION_ARG_STRING, &output_format, "output format", "FORMAT"},
 	{"dump-cert", '\0', 0, G_OPTION_ARG_NONE, &info_dumpcert, "dump certificate", NULL},
+	{"dump-recipients", '\0', 0, G_OPTION_ARG_NONE, &info_dumprecipients, "dump recipients", NULL},
 	{0}
 };
 
@@ -1915,11 +2063,34 @@ static GOptionEntry entries_service[] = {
 	{0}
 };
 
+static GOptionEntry entries_signing[] = {
+	{"cert", '\0', G_OPTION_FLAG_NOALIAS, G_OPTION_ARG_FILENAME, &certpath, "signing cert file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
+	{"key", '\0', G_OPTION_FLAG_NOALIAS, G_OPTION_ARG_FILENAME, &keypath, "signing key file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
+	{0}
+};
+
+static GOptionEntry entries_bundle_access[] = {
+	{"tls-cert", '\0', 0, G_OPTION_ARG_STRING, &access_args.tls_cert, "TLS client certificate file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
+	{"tls-key", '\0', 0, G_OPTION_ARG_STRING, &access_args.tls_key, "TLS client key file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
+	{"tls-ca", '\0', 0, G_OPTION_ARG_FILENAME, &access_args.tls_ca, "TLS CA file", "PEMFILE"},
+	{"tls-no-verify", '\0', 0, G_OPTION_ARG_NONE, &access_args.tls_no_verify, "do not verify TLS server certificate", NULL},
+	{"http-header", 'H', 0, G_OPTION_ARG_STRING_ARRAY, &access_args.http_headers, "HTTP request header (multiple uses supported)", "'HEADER: VALUE'"},
+	{0}
+};
+
+static GOptionEntry entries_encryption[] = {
+	{"to", '\0', 0, G_OPTION_ARG_FILENAME_ARRAY, &recipients, "recipient cert(s)", "PEMFILE"},
+	{0}
+};
+
 static GOptionGroup *install_group;
 static GOptionGroup *bundle_group;
 static GOptionGroup *resign_group;
 static GOptionGroup *replace_group;
 static GOptionGroup *convert_group;
+static GOptionGroup *encrypt_group;
+static GOptionGroup *extract_signature_group;
+static GOptionGroup *extract_group;
 static GOptionGroup *info_group;
 static GOptionGroup *status_group;
 static GOptionGroup *service_group;
@@ -1928,23 +2099,39 @@ static void create_option_groups(void)
 {
 	install_group = g_option_group_new("install", "Install options:", "help dummy", NULL, NULL);
 	g_option_group_add_entries(install_group, entries_install);
+	if (ENABLE_STREAMING)
+		g_option_group_add_entries(install_group, entries_bundle_access);
 
 	if (ENABLE_CREATE) {
 		bundle_group  = g_option_group_new("bundle", "Bundle options:", "help dummy", NULL, NULL);
 		g_option_group_add_entries(bundle_group, entries_bundle);
+		g_option_group_add_entries(bundle_group, entries_signing);
 
 		resign_group  = g_option_group_new("resign", "Resign options:", "help dummy", NULL, NULL);
 		g_option_group_add_entries(resign_group, entries_resign);
+		g_option_group_add_entries(resign_group, entries_signing);
 
 		replace_group  = g_option_group_new("replace-signature", "Replace signature options:", "help dummy", NULL, NULL);
 		g_option_group_add_entries(replace_group, entries_replace);
 
 		convert_group = g_option_group_new("convert", "Convert options:", "help dummy", NULL, NULL);
 		g_option_group_add_entries(convert_group, entries_convert);
+		g_option_group_add_entries(convert_group, entries_signing);
+
+		encrypt_group = g_option_group_new("encrypt", "Encryption options:", "help dummy", NULL, NULL);
+		g_option_group_add_entries(encrypt_group, entries_encryption);
+
+		extract_signature_group = g_option_group_new("extract", "Extract signature options:", "help dummy", NULL, NULL);
+		g_option_group_add_entries(extract_signature_group, entries_extract_signature);
 	}
 
+	extract_group = g_option_group_new("extract", "Extract options:", "help dummy", NULL, NULL);
+	g_option_group_add_entries(extract_group, entries_extract);
+
 	info_group    = g_option_group_new("info", "Info options:", "help dummy", NULL, NULL);
 	g_option_group_add_entries(info_group, entries_info);
+	if (ENABLE_STREAMING)
+		g_option_group_add_entries(info_group, entries_bundle_access);
 
 	status_group  = g_option_group_new("status", "Status options:", "help dummy", NULL, NULL);
 	g_option_group_add_entries(status_group, entries_status);
@@ -1956,13 +2143,14 @@ static void create_option_groups(void)
 static void cmdline_handler(int argc, char **argv)
 {
 	gboolean help = FALSE, debug = FALSE, version = FALSE;
-	gchar *confpath = NULL, *certpath = NULL, *keypath = NULL, *keyring = NULL, **intermediate = NULL, *mount = NULL;
+	gchar *confpath = NULL, *keyring = NULL, **intermediate = NULL, *mount = NULL;
 	char *cmdarg = NULL;
 	g_autoptr(GOptionContext) context = NULL;
 	GOptionEntry entries[] = {
 		{"conf", 'c', 0, G_OPTION_ARG_FILENAME, &confpath, "config file", "FILENAME"},
-		{"cert", '\0', 0, G_OPTION_ARG_FILENAME, &certpath, "cert file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
-		{"key", '\0', 0, G_OPTION_ARG_FILENAME, &keypath, "key file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
+		/* NOTE: cert and key kept for backwards-compatibility, but made invisible */
+		{"cert", '\0', G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_FILENAME, &certpath, "cert file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
+		{"key", '\0', G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_FILENAME, &keypath, "key file or PKCS#11 URL", "PEMFILE|PKCS11-URL"},
 		{"keyring", '\0', 0, G_OPTION_ARG_FILENAME, &keyring, "keyring file", "PEMFILE"},
 		{"intermediate", '\0', 0, G_OPTION_ARG_FILENAME_ARRAY, &intermediate, "intermediate CA file name", "PEMFILE"},
 		{"mount", '\0', 0, G_OPTION_ARG_FILENAME, &mount, "mount prefix", "PATH"},
@@ -1992,16 +2180,18 @@ static void cmdline_handler(int argc, char **argv)
 		{CONVERT, "convert", "convert <INBUNDLE> <OUTBUNDLE>",
 		 "Convert to casync index bundle and store",
 		 convert_start, convert_group, R_CONTEXT_CONFIG_MODE_NONE, FALSE},
+		{ENCRYPT, "encrypt", "encrypt <INBUNDLE> <OUTBUNDLE>", "Encrypt a crypt bundle",
+		 encrypt_start, encrypt_group, R_CONTEXT_CONFIG_MODE_NONE, FALSE},
 		{REPLACE_SIG, "replace-signature", "replace-signature <INBUMDLE> <INPUTSIG> <OUTBUNDLE>",
 		 "Replaces the signature of an already signed bundle",
 		 replace_signature_start, replace_group, R_CONTEXT_CONFIG_MODE_NONE, FALSE},
 		{EXTRACT_SIG, "extract-signature", "extract-signature <BUNDLENAME> <OUTPUTSIG>",
 		 "Extract the bundle signature",
-		 extract_signature_start, NULL, R_CONTEXT_CONFIG_MODE_NONE, FALSE},
+		 extract_signature_start, extract_signature_group, R_CONTEXT_CONFIG_MODE_NONE, FALSE},
 #endif
 		{EXTRACT, "extract", "extract <BUNDLENAME> <OUTPUTDIR>",
 		 "Extract the bundle content",
-		 extract_start, NULL, R_CONTEXT_CONFIG_MODE_AUTO, FALSE},
+		 extract_start, extract_group, R_CONTEXT_CONFIG_MODE_AUTO, FALSE},
 		{INFO, "info", "info <FILE>",
 		 "Print bundle info",
 		 info_start, info_group, R_CONTEXT_CONFIG_MODE_AUTO, FALSE},
@@ -2041,12 +2231,13 @@ static void cmdline_handler(int argc, char **argv)
 			"  bundle\t\tCreate a bundle\n"
 			"  resign\t\tResign an already signed bundle\n"
 			"  convert\t\tConvert classic to casync bundle\n"
+			"  encrypt\t\tEncrypt a crypt bundle\n"
 			"  replace-signature\tReplaces the signature of an already signed bundle\n"
 			"  extract-signature\tExtract the bundle signature\n"
 #endif
 			"  extract\t\tExtract the bundle content\n"
 			"  install\t\tInstall a bundle\n"
-			"  info\t\t\tShow file information\n"
+			"  info\t\t\tShow bundle information\n"
 			"  mount\t\t\tMount a bundle\n"
 #if ENABLE_SERVICE == 1
 			"  service\t\tStart RAUC service\n"
@@ -2112,7 +2303,7 @@ static void cmdline_handler(int argc, char **argv)
 
 	if (rcommand == NULL) {
 		/* INVALID COMMAND given */
-		g_message("Invalid command '%s' given", cmdarg);
+		g_print("Invalid command '%s' given\n", cmdarg);
 		r_exit_status = 1;
 		goto print_help;
 	}
@@ -2153,8 +2344,14 @@ static void cmdline_handler(int argc, char **argv)
 			r_context_conf()->configpath = confpath;
 		if (certpath)
 			r_context_conf()->certpath = certpath;
-		if (keypath)
-			r_context_conf()->keypath = keypath;
+		if (keypath) {
+			/* 'key' means encryption key for 'info', 'extract' or 'extract-signature',
+			 * signing key otherwise */
+			if (rcommand->type == INFO || rcommand->type == EXTRACT || rcommand->type == EXTRACT_SIG)
+				r_context_conf()->encryption_key = keypath;
+			else
+				r_context_conf()->keypath = keypath;
+		}
 		if (keyring)
 			r_context_conf()->keyringpath = keyring;
 		if (signing_keyring)
@@ -2163,6 +2360,8 @@ static void cmdline_handler(int argc, char **argv)
 			r_context_conf()->mksquashfs_args = mksquashfs_args;
 		if (casync_args)
 			r_context_conf()->casync_args = casync_args;
+		if (recipients)
+			r_context_conf()->recipients = recipients;
 		if (intermediate)
 			r_context_conf()->intermediatepaths = intermediate;
 		if (mount)
@@ -2209,6 +2408,14 @@ int main(int argc, char **argv)
 {
 	GLogLevelFlags fatal_mask;
 
+#if GLIB_CHECK_VERSION(2, 68, 0)
+	/* To use this function, without bumping the maximum GLib allowed version,
+	 * we temporarily disable the deprecation warnings */
+	G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+	g_log_writer_default_set_use_stderr(TRUE);
+	G_GNUC_END_IGNORE_DEPRECATIONS
+#endif
+
 	fatal_mask = g_log_set_always_fatal(G_LOG_FATAL_MASK);
 	fatal_mask |= G_LOG_LEVEL_CRITICAL;
 	g_log_set_always_fatal(fatal_mask);
@@ -2221,6 +2428,21 @@ int main(int argc, char **argv)
 	if (g_get_charset(NULL))
 		utf8_supported = TRUE;
 
+	if (ENABLE_STREAMING && g_getenv("RAUC_NBD_SERVER")) {
+		GError *ierror = NULL;
+		pthread_setname_np(pthread_self(), "rauc-nbd");
+		if (r_nbd_run_server(RAUC_SOCKET_FD, &ierror)) {
+			return 0;
+		} else {
+			if (ierror) {
+				g_message("nbd server failed with: %s", ierror->message);
+			} else {
+				g_message("nbd server failed");
+			}
+			return 1;
+		}
+	}
+
 	create_option_groups();
 	cmdline_handler(argc, argv);
 
