@@ -78,6 +78,9 @@ static gboolean parse_image(GKeyFile *key_file, const gchar *group, RaucImage **
 		goto out;
 	}
 
+	iimage->adaptive = g_key_file_get_string_list(key_file, group, "adaptive", NULL, NULL);
+	g_key_file_remove_key(key_file, group, "adaptive", NULL);
+
 	if (!check_remaining_keys(key_file, group, &ierror)) {
 		g_propagate_error(error, ierror);
 		goto out;
@@ -104,28 +107,29 @@ out:
 static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GError **error)
 {
 	GError *ierror = NULL;
-	RaucManifest *raucm = g_new0(RaucManifest, 1);
-	gboolean res = FALSE;
+	g_autoptr(RaucManifest) raucm = g_new0(RaucManifest, 1);
 	g_autofree gchar *tmp = NULL;
-	gchar **groups;
+	g_auto(GStrv) groups = NULL;
 	gsize group_count;
 	g_auto(GStrv) bundle_hooks = NULL;
 	gsize hook_entries;
 
-	g_assert_null(*manifest);
+	g_return_val_if_fail(key_file != NULL, FALSE);
+	g_return_val_if_fail(manifest != NULL && *manifest == NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
 	/* parse [update] section */
 	raucm->update_compatible = key_file_consume_string(key_file, "update", "compatible", &ierror);
 	if (!raucm->update_compatible) {
 		g_propagate_error(error, ierror);
-		goto free;
+		return FALSE;
 	}
 	raucm->update_version = key_file_consume_string(key_file, "update", "version", NULL);
 	raucm->update_description = key_file_consume_string(key_file, "update", "description", NULL);
 	raucm->update_build = key_file_consume_string(key_file, "update", "build", NULL);
 	if (!check_remaining_keys(key_file, "update", &ierror)) {
 		g_propagate_error(error, ierror);
-		goto free;
+		return FALSE;
 	}
 	g_key_file_remove_group(key_file, "update", NULL);
 
@@ -133,9 +137,9 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 	tmp = key_file_consume_string(key_file, "bundle", "format", NULL);
 	if (tmp == NULL || g_strcmp0(tmp, "plain") == 0) {
 		raucm->bundle_format = R_MANIFEST_FORMAT_PLAIN;
-	} else if (g_strcmp0(tmp, "verity") == 0) {
+	} else if ((g_strcmp0(tmp, "verity") == 0) || (g_strcmp0(tmp, "crypt") == 0)) {
 		/* only SHA256 is supported for now */
-		raucm->bundle_format = R_MANIFEST_FORMAT_VERITY;
+		raucm->bundle_format = g_strcmp0(tmp, "crypt") == 0 ? R_MANIFEST_FORMAT_CRYPT : R_MANIFEST_FORMAT_VERITY;
 		raucm->bundle_verity_hash = key_file_consume_string(key_file, "bundle", "verity-hash", NULL);
 		raucm->bundle_verity_salt = key_file_consume_string(key_file, "bundle", "verity-salt", NULL);
 		raucm->bundle_verity_size = g_key_file_get_uint64(key_file, "bundle", "verity-size", NULL);
@@ -144,11 +148,15 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 	} else {
 		g_set_error(error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_PARSE,
 				"Invalid format value '%s' in group '[bundle]'", tmp);
-		goto free;
+		return FALSE;
+	}
+	/* crypt format requires additional dm-crypt key */
+	if (g_strcmp0(tmp, "crypt") == 0) {
+		raucm->bundle_crypt_key = key_file_consume_string(key_file, "bundle", "crypt-key", NULL);
 	}
 	if (!check_remaining_keys(key_file, "bundle", &ierror)) {
 		g_propagate_error(error, ierror);
-		goto free;
+		return FALSE;
 	}
 	g_key_file_remove_group(key_file, "bundle", NULL);
 
@@ -157,7 +165,7 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 	raucm->handler_args = key_file_consume_string(key_file, "handler", "args", NULL);
 	if (!check_remaining_keys(key_file, "handler", &ierror)) {
 		g_propagate_error(error, ierror);
-		goto free;
+		return FALSE;
 	}
 	g_key_file_remove_group(key_file, "handler", NULL);
 
@@ -171,42 +179,43 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 		} else {
 			g_set_error(error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_PARSE,
 					"install hook type '%s' not supported", bundle_hooks[j]);
-			goto free;
+			return FALSE;
 		}
 	}
 
 	if (!check_remaining_keys(key_file, "hooks", &ierror)) {
 		g_propagate_error(error, ierror);
-		goto free;
+		return FALSE;
 	}
 	g_key_file_remove_group(key_file, "hooks", NULL);
 
-	/* parse [image.<slotclass>] sections */
 	groups = g_key_file_get_groups(key_file, &group_count);
 	for (gsize i = 0; i < group_count; i++) {
+		/* parse [image.<slotclass>] sections */
 		if (g_str_has_prefix(groups[i], RAUC_IMAGE_PREFIX ".")) {
 			RaucImage *image = NULL;
 
 			if (!parse_image(key_file, groups[i], &image, &ierror)) {
 				g_propagate_error(error, ierror);
-				goto free;
+				return FALSE;
 			}
 
 			raucm->images = g_list_append(raucm->images, image);
 		}
+		/* ignore [meta.<label>] sections */
+		if (g_str_has_prefix(groups[i], "meta.")) {
+			g_key_file_remove_group(key_file, groups[i], NULL);
+		}
 	}
 
 	if (!check_remaining_groups(key_file, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto free;
+		return FALSE;
 	}
 
-	g_strfreev(groups);
-
-	res = TRUE;
 	*manifest = g_steal_pointer(&raucm);
-free:
-	return res;
+
+	return TRUE;
 }
 
 gboolean load_manifest_mem(GBytes *mem, RaucManifest **manifest, GError **error)
@@ -274,6 +283,7 @@ static gboolean check_manifest_common(const RaucManifest *mf, GError **error)
 		case R_MANIFEST_FORMAT_PLAIN:
 			break; /* no additional data needed */
 		case R_MANIFEST_FORMAT_VERITY:
+		case R_MANIFEST_FORMAT_CRYPT:
 			break; /* data checked in _detached/_inline */
 		default: {
 			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unsupported bundle format");
@@ -362,17 +372,24 @@ gboolean check_manifest_internal(const RaucManifest *mf, GError **error)
 	switch (mf->bundle_format) {
 		case R_MANIFEST_FORMAT_PLAIN:
 			break; /* no additional data needed */
+		case R_MANIFEST_FORMAT_CRYPT: {
+			if (mf->bundle_crypt_key) {
+				g_set_error_literal(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected key for crypt bundle in internal manifest");
+				goto out;
+			}
+		};
+		/* Fallthrough */
 		case R_MANIFEST_FORMAT_VERITY: {
 			if (mf->bundle_verity_hash) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected hash for verity bundle in internal manifest");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected hash for %s bundle in internal manifest", r_manifest_bundle_format_to_str(mf->bundle_format));
 				goto out;
 			}
 			if (mf->bundle_verity_salt) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected hash for verity bundle in internal manifest");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected hash for %s bundle in internal manifest", r_manifest_bundle_format_to_str(mf->bundle_format));
 				goto out;
 			}
 			if (mf->bundle_verity_size) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected hash for verity bundle in internal manifest");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unexpected hash for %s bundle in internal manifest", r_manifest_bundle_format_to_str(mf->bundle_format));
 				goto out;
 			}
 
@@ -394,9 +411,16 @@ gboolean check_manifest_external(const RaucManifest *mf, GError **error)
 {
 	GError *ierror = NULL;
 	gboolean res = FALSE;
+	const gchar *format = NULL;
 
 	r_context_begin_step("check_manifest", "Checking manifest contents", 0);
 
+	format = r_manifest_bundle_format_to_str(mf->bundle_format);
+	if (g_strcmp0(format, "invalid") == 0) {
+		g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unsupported bundle format");
+		goto out;
+	}
+
 	if (!check_manifest_common(mf, &ierror)) {
 		g_propagate_error(error, ierror);
 		goto out;
@@ -407,45 +431,61 @@ gboolean check_manifest_external(const RaucManifest *mf, GError **error)
 			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unsupported bundle format 'plain' for external manifest");
 			goto out;
 		}
+		case R_MANIFEST_FORMAT_CRYPT: {
+			guint8 *tmp;
+
+			if (!mf->bundle_crypt_key) {
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing key for crypt bundle");
+				goto out;
+			}
+			tmp = r_hex_decode(mf->bundle_crypt_key, 32);
+			if (!tmp) {
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Invalid key for crypt bundle");
+				goto out;
+			}
+			g_free(tmp);
+		};
+		/* Fallthrough */
 		case R_MANIFEST_FORMAT_VERITY: {
 			guint8 *tmp;
 
 			if (!mf->bundle_verity_hash) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing hash for verity bundle");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing hash for %s bundle", format);
 				goto out;
 			}
 			tmp = r_hex_decode(mf->bundle_verity_hash, 32);
 			if (!tmp) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Invalid hash for verity bundle");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Invalid hash for %s bundle", format);
 				goto out;
 			}
 			g_free(tmp);
 
 			if (!mf->bundle_verity_salt) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing salt for verity bundle");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing salt for %s bundle", format);
 				goto out;
 			}
 			tmp = r_hex_decode(mf->bundle_verity_salt, 32);
 			if (!tmp) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Invalid salt for verity bundle");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Invalid salt for %s bundle", format);
 				goto out;
 			}
 			g_free(tmp);
 
 			if (!mf->bundle_verity_size) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing size for verity bundle");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Missing size for %s bundle", format);
 				goto out;
 			}
 
 			if (mf->bundle_verity_size % 4096) {
-				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unaligned size for verity bundle");
+				g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unaligned size for %s bundle", format);
 				goto out;
 			}
 
 			break;
 		};
 		default: {
-			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unsupported bundle format");
+			/* should not be reached as this is checked before */
+			g_error("Unsupported bundle format");
 			goto out;
 		}
 	}
@@ -478,8 +518,13 @@ static GKeyFile *prepare_manifest(const RaucManifest *mf)
 	switch (mf->bundle_format) {
 		case R_MANIFEST_FORMAT_PLAIN:
 			break;
+		case R_MANIFEST_FORMAT_CRYPT: {
+			if (mf->bundle_crypt_key)
+				g_key_file_set_string(key_file, "bundle", "crypt-key", mf->bundle_crypt_key);
+		};
+		/* Fallthrough */
 		case R_MANIFEST_FORMAT_VERITY: {
-			g_key_file_set_string(key_file, "bundle", "format", "verity");
+			g_key_file_set_string(key_file, "bundle", "format", r_manifest_bundle_format_to_str(mf->bundle_format));
 			if (mf->bundle_verity_hash)
 				g_key_file_set_string(key_file, "bundle", "verity-hash", mf->bundle_verity_hash);
 			if (mf->bundle_verity_salt)
@@ -551,6 +596,10 @@ static GKeyFile *prepare_manifest(const RaucManifest *mf)
 			g_key_file_set_string_list(key_file, group, "hooks",
 					(const gchar **)hooklist->pdata, hooklist->len);
 		}
+
+		if (image->adaptive)
+			g_key_file_set_string_list(key_file, group, "adaptive",
+					(const gchar * const *)image->adaptive, g_strv_length(image->adaptive));
 	}
 
 	return g_steal_pointer(&key_file);
@@ -572,7 +621,7 @@ gboolean save_manifest_mem(GBytes **mem, const RaucManifest *mf)
 	g_assert(data != NULL);
 	g_assert(length > 0);
 
-	*mem = g_bytes_new(data, length);
+	*mem = g_bytes_new_take(data, length);
 
 	return TRUE;
 }
@@ -603,6 +652,7 @@ void r_free_image(gpointer data)
 	g_free(image->variant);
 	g_free(image->checksum.digest);
 	g_free(image->filename);
+	g_strfreev(image->adaptive);
 	g_free(image);
 }
 
