@@ -19,6 +19,10 @@ gboolean r_mount_bundle(const gchar *source, const gchar *mountpoint, GError **e
 {
 	const unsigned long flags = MS_NODEV | MS_NOSUID | MS_RDONLY;
 
+	g_return_val_if_fail(source != NULL, FALSE);
+	g_return_val_if_fail(mountpoint != NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	if (mount(source, mountpoint, "squashfs", flags, NULL)) {
 		int err = errno;
 		g_set_error(error,
@@ -35,6 +39,9 @@ gboolean r_umount_bundle(const gchar *mountpoint, GError **error)
 {
 	const int flags = UMOUNT_NOFOLLOW;
 
+	g_return_val_if_fail(mountpoint != NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	if (umount2(mountpoint, flags)) {
 		int err = errno;
 		g_set_error(error,
@@ -58,10 +65,6 @@ gboolean r_mount_full(const gchar *source, const gchar *mountpoint, const gchar*
 	g_return_val_if_fail(mountpoint != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	if (getuid() != 0) {
-		g_ptr_array_add(args, g_strdup("sudo"));
-		g_ptr_array_add(args, g_strdup("--non-interactive"));
-	}
 	g_ptr_array_add(args, g_strdup("mount"));
 	if (type != NULL) {
 		g_ptr_array_add(args, g_strdup("-t"));
@@ -138,8 +141,10 @@ gboolean r_setup_loop(gint fd, gint *loopfd_out, gchar **loopname_out, goffset s
 		gint loopidx;
 
 		g_clear_pointer(&loopname, g_free);
-		if (loopfd >= 0)
+		if (loopfd >= 0) {
 			g_close(loopfd, NULL);
+			loopfd = -1;
+		}
 
 		loopidx = ioctl(controlfd, LOOP_CTL_GET_FREE);
 		if (loopidx < 0) {
@@ -211,9 +216,12 @@ gboolean r_setup_loop(gint fd, gint *loopfd_out, gchar **loopname_out, goffset s
 		goto out;
 	}
 
-	looprc = ioctl(loopfd, LOOP_SET_BLOCK_SIZE, 4096);
+	do {
+		looprc = ioctl(loopfd, LOOP_SET_BLOCK_SIZE, 4096);
+	} while (looprc < 0 && errno == EAGAIN);
 	if (looprc < 0) {
-		g_debug("Failed to set loop device block size to 4096, continuing");
+		g_warning("Failed to set loop device block size to 4096: %s, continuing",
+				g_strerror(errno));
 	}
 
 	g_message("Configured loop device '%s' for %" G_GOFFSET_FORMAT " bytes", loopname, size);
@@ -241,10 +249,6 @@ gboolean r_umount(const gchar *filename, GError **error)
 	g_return_val_if_fail(filename != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	if (getuid() != 0) {
-		g_ptr_array_add(args, g_strdup("sudo"));
-		g_ptr_array_add(args, g_strdup("--non-interactive"));
-	}
 	g_ptr_array_add(args, g_strdup("umount"));
 	g_ptr_array_add(args, g_strdup(filename));
 	g_ptr_array_add(args, NULL);
@@ -290,12 +294,14 @@ gchar* r_create_mount_point(const gchar *name, GError **error)
 		ret = g_mkdir_with_parents(mountpoint, 0700);
 
 		if (ret != 0) {
+			int err = errno;
 			g_set_error(
 					error,
 					G_FILE_ERROR,
-					G_FILE_ERROR_FAILED,
-					"Failed creating mount path '%s'",
-					mountpoint);
+					g_file_error_from_errno(err),
+					"Failed creating mount path '%s': %s",
+					mountpoint,
+					g_strerror(err));
 			g_free(mountpoint);
 			mountpoint = NULL;
 			goto out;
