@@ -17,6 +17,21 @@ GQuark r_signature_error_quark(void)
 	return g_quark_from_static_string("r_signature_error_quark");
 }
 
+static const gchar *get_openssl_err_string(void)
+{
+	unsigned long err;
+	const gchar *data = NULL;
+	int errflags = 0;
+
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+	err = ERR_get_error_line_data(NULL, NULL, &data, &errflags);
+#else
+	err = ERR_get_error_all(NULL, NULL, NULL, &data, &errflags);
+#endif
+
+	return (errflags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL);
+}
+
 /* return 0 for error, 1 for success */
 static int check_purpose_code_sign(const X509_PURPOSE *xp, const X509 *const_x, int ca)
 {
@@ -62,17 +77,11 @@ gboolean signature_init(GError **error)
 
 	ret = OPENSSL_init_crypto(OPENSSL_INIT_LOAD_CONFIG | OPENSSL_INIT_LOAD_CRYPTO_STRINGS, NULL);
 	if (!ret) {
-		unsigned long err;
-		const gchar *data;
-		int flags;
-
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_CRYPTOINIT_FAILED,
-				"Failed to initialize OpenSSL crypto: %s",
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"Failed to initialize OpenSSL crypto: %s", get_openssl_err_string());
 		return FALSE;
 	}
 
@@ -89,17 +98,11 @@ gboolean signature_init(GError **error)
 	/* X509_TRUST_OBJECT_SIGN maps to the Code Signing ID (via OpenSSL's NID_code_sign) */
 	ret = X509_PURPOSE_add(id, X509_TRUST_OBJECT_SIGN, 0, check_purpose_code_sign, "Code signing", "codesign", NULL);
 	if (!ret) {
-		unsigned long err;
-		const gchar *data;
-		int flags;
-
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_CRYPTOINIT_FAILED,
-				"Failed to configure OpenSSL X509 purpose: %s",
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"Failed to configure OpenSSL X509 purpose: %s", get_openssl_err_string());
 		return FALSE;
 	}
 
@@ -109,10 +112,7 @@ gboolean signature_init(GError **error)
 static ENGINE *get_pkcs11_engine(GError **error)
 {
 	static ENGINE *e = NULL;
-	unsigned long err;
-	const gchar *data;
 	const gchar *env;
-	int flags;
 
 	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
 
@@ -120,13 +120,11 @@ static ENGINE *get_pkcs11_engine(GError **error)
 
 	e = ENGINE_by_id("pkcs11");
 	if (e == NULL) {
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_LOAD_FAILED,
-				"failed to load PKCS11 engine: %s",
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"failed to load PKCS11 engine: %s", get_openssl_err_string());
 
 		goto out;
 	}
@@ -134,25 +132,21 @@ static ENGINE *get_pkcs11_engine(GError **error)
 	env = g_getenv("RAUC_PKCS11_MODULE");
 	if (env != NULL) {
 		if (!ENGINE_ctrl_cmd_string(e, "MODULE_PATH", env, 0)) {
-			err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 			g_set_error(
 					error,
 					R_SIGNATURE_ERROR,
 					R_SIGNATURE_ERROR_PARSE_ERROR,
-					"failed to configure PKCS11 module path: %s",
-					(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+					"failed to configure PKCS11 module path: %s", get_openssl_err_string());
 			goto free;
 		}
 	}
 
 	if (ENGINE_init(e) == 0) {
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_LOAD_FAILED,
-				"failed to initialize PKCS11 engine: %s",
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"failed to initialize PKCS11 engine: %s", get_openssl_err_string());
 
 		goto free;
 	}
@@ -160,13 +154,11 @@ static ENGINE *get_pkcs11_engine(GError **error)
 	env = g_getenv("RAUC_PKCS11_PIN");
 	if (env != NULL && env[0] != '\0') {
 		if (!ENGINE_ctrl_cmd_string(e, "PIN", env, 0)) {
-			err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 			g_set_error(
 					error,
 					R_SIGNATURE_ERROR,
 					R_SIGNATURE_ERROR_PARSE_ERROR,
-					"failed to configure PKCS11 PIN: %s",
-					(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+					"failed to configure PKCS11 PIN: %s", get_openssl_err_string());
 			goto finish;
 		}
 	}
@@ -186,9 +178,7 @@ static EVP_PKEY *load_key_file(const gchar *keyfile, GError **error)
 {
 	EVP_PKEY *res = NULL;
 	BIO *key = NULL;
-	unsigned long err;
-	const gchar *data;
-	int flags;
+	const gchar *passphrase;
 
 	g_return_val_if_fail(keyfile != NULL, NULL);
 	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
@@ -203,15 +193,16 @@ static EVP_PKEY *load_key_file(const gchar *keyfile, GError **error)
 		goto out;
 	}
 
-	res = PEM_read_bio_PrivateKey(key, NULL, NULL, NULL);
+	passphrase = g_getenv("RAUC_KEY_PASSPHRASE");
+	if (passphrase && passphrase[0] == '\0')
+		passphrase = NULL;
+	res = PEM_read_bio_PrivateKey(key, NULL, NULL, (void *)passphrase);
 	if (res == NULL) {
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_PARSE_ERROR,
-				"failed to parse key file '%s': %s", keyfile,
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"failed to parse key file '%s': %s", keyfile, get_openssl_err_string());
 		goto out;
 	}
 out:
@@ -222,10 +213,7 @@ out:
 static EVP_PKEY *load_key_pkcs11(const gchar *url, GError **error)
 {
 	EVP_PKEY *res = NULL;
-	unsigned long err;
-	const gchar *data;
 	GError *ierror = NULL;
-	int flags;
 	ENGINE *e;
 
 	g_return_val_if_fail(url != NULL, NULL);
@@ -239,13 +227,11 @@ static EVP_PKEY *load_key_pkcs11(const gchar *url, GError **error)
 
 	res = ENGINE_load_private_key(e, url, NULL, NULL);
 	if (res == NULL) {
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_LOAD_FAILED,
-				"failed to load PKCS11 private key for '%s': %s", url,
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"failed to load PKCS11 private key for '%s': %s", url, get_openssl_err_string());
 		goto out;
 	}
 out:
@@ -266,9 +252,6 @@ static X509 *load_cert_file(const gchar *certfile, GError **error)
 {
 	X509 *res = NULL;
 	BIO *cert = NULL;
-	unsigned long err;
-	const gchar *data;
-	int flags;
 
 	g_return_val_if_fail(certfile != NULL, NULL);
 	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
@@ -285,13 +268,11 @@ static X509 *load_cert_file(const gchar *certfile, GError **error)
 
 	res = PEM_read_bio_X509(cert, NULL, NULL, NULL);
 	if (res == NULL) {
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_PARSE_ERROR,
-				"failed to parse cert file '%s': %s", certfile,
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"failed to parse cert file '%s': %s", certfile, get_openssl_err_string());
 		goto out;
 	}
 
@@ -300,13 +281,62 @@ out:
 	return res;
 }
 
+static STACK_OF(X509) *load_certs_from_file(const gchar *certfile, GError **error)
+{
+	BIO *cert_bio = NULL;
+	X509 *cert_x509 = NULL;
+	STACK_OF(X509) *certs = NULL;
+	unsigned long err;
+
+	g_return_val_if_fail(certfile != NULL, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	cert_bio = BIO_new_file(certfile, "r");
+	if (cert_bio == NULL) {
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_LOAD_FAILED,
+				"Failed to load cert file '%s'", certfile);
+		goto out;
+	}
+
+	certs = sk_X509_new_null();
+
+	for (;;) {
+		cert_x509 = PEM_read_bio_X509(cert_bio, NULL, NULL, NULL);
+		if (cert_x509 == NULL) {
+			err = ERR_peek_last_error();
+
+			if (ERR_GET_REASON(err) == PEM_R_NO_START_LINE) {
+				/* simply reached end of file */
+				ERR_clear_error();
+				break;
+			}
+
+			g_set_error(
+					error,
+					R_SIGNATURE_ERROR,
+					R_SIGNATURE_ERROR_PARSE_ERROR,
+					"Failed to parse cert file '%s': %s", certfile, get_openssl_err_string());
+			/* other certs loaded so far are not required anymore and must be freed */
+			sk_X509_pop_free(certs, X509_free);
+			certs = NULL;
+			goto out;
+		}
+
+		sk_X509_push(certs, cert_x509);
+	}
+
+out:
+	BIO_free_all(cert_bio);
+	return certs;
+}
+
 static X509 *load_cert_pkcs11(const gchar *url, GError **error)
 {
 	X509 *res = NULL;
-	unsigned long err;
-	const gchar *data;
 	GError *ierror = NULL;
-	int flags;
 	ENGINE *e;
 
 	/* this is defined in libp11 src/eng_back.c ctx_ctrl_load_cert() */
@@ -327,13 +357,11 @@ static X509 *load_cert_pkcs11(const gchar *url, GError **error)
 	parms.url = url;
 	parms.cert = NULL;
 	if (!ENGINE_ctrl_cmd(e, "LOAD_CERT_CTRL", 0, &parms, NULL, 0) || (parms.cert == NULL)) {
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_PARSE_ERROR,
-				"failed to load PKCS11 certificate for '%s': %s", url,
-				(flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"failed to load PKCS11 certificate for '%s': %s", url, get_openssl_err_string());
 		goto out;
 	}
 	res = parms.cert;
@@ -438,6 +466,10 @@ X509_STORE* setup_x509_store(const gchar *capath, const gchar *cadir, GError **e
 	else if (contains_crl(load_capath, load_cadir))
 		g_warning("Detected CRL but CRL checking is disabled!");
 
+	/* Allow partial chain if configured */
+	if (r_context()->config->keyring_allow_partial_chain)
+		X509_STORE_set_flags(store, X509_V_FLAG_PARTIAL_CHAIN);
+
 	/* Enable purpose checking if configured */
 	if (check_purpose) {
 		const X509_PURPOSE *xp = X509_PURPOSE_get0(X509_PURPOSE_get_by_sname(check_purpose));
@@ -460,10 +492,10 @@ GBytes *cms_sign(GBytes *content, gboolean detached, const gchar *certfile, cons
 	BIO *incontent = BIO_new_mem_buf((void *)g_bytes_get_data(content, NULL),
 			g_bytes_get_size(content));
 	BIO *outsig = BIO_new(BIO_s_mem());
-	X509 *signcert = NULL;
-	EVP_PKEY *pkey = NULL;
+	g_autoptr(X509) signcert = NULL;
+	g_autoptr(EVP_PKEY) pkey = NULL;
 	STACK_OF(X509) *intercerts = NULL;
-	CMS_ContentInfo *cms = NULL;
+	g_autoptr(CMS_ContentInfo) cms = NULL;
 	GBytes *res = NULL;
 	int flags = CMS_BINARY | CMS_NOSMIMECAP;
 	const gchar *keyring_path = NULL, *keyring_dir = NULL;
@@ -502,15 +534,11 @@ GBytes *cms_sign(GBytes *content, gboolean detached, const gchar *certfile, cons
 
 	cms = CMS_sign(signcert, pkey, intercerts, incontent, flags);
 	if (cms == NULL) {
-		unsigned long err;
-		const gchar *data;
-		int errflags;
-		err = ERR_get_error_line_data(NULL, NULL, &data, &errflags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
-				R_SIGNATURE_ERROR_INVALID,
-				"failed to create signature: %s", (errflags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				R_SIGNATURE_ERROR_CREATE_SIG,
+				"failed to create signature: %s", get_openssl_err_string());
 		goto out;
 	}
 	if (!i2d_CMS_bio(outsig, cms)) {
@@ -596,6 +624,8 @@ GBytes *cms_sign(GBytes *content, gboolean detached, const gchar *certfile, cons
 	}
 out:
 	ERR_print_errors_fp(stdout);
+	if (intercerts)
+		sk_X509_pop_free(intercerts, X509_free);
 	BIO_free_all(incontent);
 	BIO_free_all(outsig);
 	return res;
@@ -603,8 +633,7 @@ out:
 
 gchar* get_pubkey_hash(X509 *cert)
 {
-	gchar *data = NULL;
-	GString *string;
+	g_autoptr(GString) string = NULL;
 	g_autofree unsigned char *der_buf = NULL;
 	unsigned char *tmp_buf = NULL;
 	unsigned int len = 0;
@@ -619,7 +648,7 @@ gchar* get_pubkey_hash(X509 *cert)
 	len = i2d_X509_PUBKEY(X509_get_X509_PUBKEY(cert), NULL);
 	if (len <= 0) {
 		g_warning("DER Encoding failed");
-		goto out;
+		return NULL;
 	}
 	/* As i2d_X509_PUBKEY() moves pointer after end of data,
 	 * we must use a tmp pointer, here */
@@ -630,7 +659,7 @@ gchar* get_pubkey_hash(X509 *cert)
 
 	if (!EVP_Digest(der_buf, len, md, &n, EVP_sha256(), NULL)) {
 		g_warning("Error in EVP_Digest");
-		goto out;
+		return NULL;
 	}
 
 	g_assert_cmpint(n, ==, SHA256_DIGEST_LENGTH);
@@ -640,9 +669,7 @@ gchar* get_pubkey_hash(X509 *cert)
 	}
 	g_string_truncate(string, SHA256_DIGEST_LENGTH * 3 - 1);
 
-	data = g_string_free(string, FALSE);
-out:
-	return data;
+	return g_string_free(g_steal_pointer(&string), FALSE);
 }
 
 gchar** get_pubkey_hashes(STACK_OF(X509) *verified_chain)
@@ -669,29 +696,42 @@ out:
 	return ret;
 }
 
+/*
+ * Reads text out of BIO.
+ *
+ * @param input BIO, will be freed
+ *
+ * @return newly allocated string or NULL
+ */
+static gchar* bio_mem_unwrap(BIO *mem)
+{
+	long size;
+	gchar *data, *ret;
+
+	g_return_val_if_fail(mem != NULL, NULL);
+
+	size = BIO_get_mem_data(mem, &data);
+	ret = g_strndup(data, size);
+	BIO_free(mem);
+
+	return ret;
+}
+
 static gchar* dump_cms(STACK_OF(X509) *x509_certs)
 {
 	BIO *mem;
-	gchar *data, *ret;
-	gsize size;
 
 	g_return_val_if_fail(x509_certs != NULL, NULL);
 
 	mem = BIO_new(BIO_s_mem());
 	X509_print_ex(mem, sk_X509_value(x509_certs, 0), 0, 0);
 
-	size = BIO_get_mem_data(mem, &data);
-	ret = g_strndup(data, size);
-
-	BIO_set_close(mem, BIO_CLOSE);
-	BIO_free(mem);
-
-	return ret;
+	return bio_mem_unwrap(mem);
 }
 
 gchar* sigdata_to_string(GBytes *sig, GError **error)
 {
-	CMS_ContentInfo *cms = NULL;
+	g_autoptr(CMS_ContentInfo) cms = NULL;
 	STACK_OF(X509) *signers = NULL;
 	gchar *ret;
 	BIO *insig = BIO_new_mem_buf((void *)g_bytes_get_data(sig, NULL),
@@ -721,7 +761,137 @@ gchar* sigdata_to_string(GBytes *sig, GError **error)
 
 	ret = dump_cms(signers);
 
-	sk_X509_free(signers);
+	sk_X509_pop_free(signers, X509_free);
+	BIO_free(insig);
+
+	return ret;
+}
+
+static void bio_print_recipient(BIO *text, guint id, gchar* algorithm, ASN1_OCTET_STRING *keyid, X509_NAME *issuer, ASN1_INTEGER *sno)
+{
+	g_autofree gchar *s = NULL;
+
+	/* OpenSSL documentation says
+	 * "Either the keyidentifier will be set in keyid or both
+	 * issuer name and serial number in issuer and sno."
+	 */
+	if (keyid) {
+		/* dummy printout first of all */
+		BIO_printf(text, "%3d   <keyid>", id);
+		return;
+	}
+
+	if (!issuer || !sno) {
+		BIO_printf(text, "%3d   <unknown>", id);
+		return;
+	}
+
+	BIO_printf(text, "%3d   Issuer:    ", id);
+	X509_NAME_print_ex(text, issuer, 0, XN_FLAG_ONELINE);
+	BIO_puts(text, "\n");
+	BIO_puts(text, "      Serial:    ");
+	s = i2s_ASN1_INTEGER(NULL, sno);
+	BIO_puts(text, s);
+	BIO_puts(text, "\n");
+	if (algorithm)
+		BIO_printf(text, "      Algorithm: %s\n", algorithm);
+}
+
+gchar* envelopeddata_to_string(GBytes *sig, GError **error)
+{
+	g_autoptr(CMS_ContentInfo) cms = NULL;
+	BIO *insig = NULL;
+	STACK_OF(CMS_RecipientInfo) *ris;
+	BIO *text;
+	gchar *ret;
+
+	g_return_val_if_fail(sig != NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	insig = BIO_new_mem_buf((void *)g_bytes_get_data(sig, NULL),
+			g_bytes_get_size(sig));
+	if (!insig)
+		g_error("BIO_new_mem_buf() failed");
+
+	if (!(cms = d2i_CMS_bio(insig, NULL))) {
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_PARSE,
+				"Failed to parse signature");
+		return NULL;
+	}
+	ris = CMS_get0_RecipientInfos(cms);
+
+	text = BIO_new(BIO_s_mem());
+	BIO_printf(text, "%d Recipients:\n", sk_CMS_RecipientInfo_num(ris));
+
+	for (int i = 0; i < sk_CMS_RecipientInfo_num(ris); i++) {
+		CMS_RecipientInfo *ri;
+
+		ri = sk_CMS_RecipientInfo_value(ris, i);
+
+		switch (CMS_RecipientInfo_type(ri)) {
+			case CMS_RECIPINFO_TRANS: {
+				ASN1_OCTET_STRING *keyid = NULL;
+				X509_NAME *issuer = NULL;
+				ASN1_INTEGER *sno = NULL;
+				X509_ALGOR *alg = NULL;
+				gchar algo_buf[80];
+
+				if (CMS_RecipientInfo_ktri_get0_signer_id(ri, &keyid, &issuer, &sno) != 1) {
+					g_warning("Unable to obtain recipient information for recipient %d", i);
+				}
+
+				if (CMS_RecipientInfo_ktri_get0_algs(ri, NULL, NULL, &alg) != 1) {
+					g_warning("Unable to obtain algorithm information for recipient %d", i);
+				}
+
+				OBJ_obj2txt(algo_buf, sizeof(algo_buf), alg->algorithm, 0);
+
+				bio_print_recipient(text, i, algo_buf, keyid, issuer, sno);
+			} break;
+			case CMS_RECIPINFO_AGREE: {
+				STACK_OF(CMS_RecipientEncryptedKey) *reks = NULL;
+				X509_ALGOR *alg = NULL;
+				gchar algo_buf[80];
+
+				reks = CMS_RecipientInfo_kari_get0_reks(ri);
+				if (!reks) {
+					g_warning("Unable to obtain recipient information for recipient %d", i);
+				}
+
+				if (CMS_RecipientInfo_kari_get0_alg(ri, &alg, NULL) != 1) {
+					g_warning("Unable to obtain algorithm information for recipient %d", i);
+				}
+
+				OBJ_obj2txt(algo_buf, sizeof(algo_buf), alg->algorithm, 0);
+
+				for (int j = 0; j < sk_CMS_RecipientEncryptedKey_num(reks); j++) {
+					ASN1_OCTET_STRING *keyid = NULL;
+					X509_NAME *issuer = NULL;
+					ASN1_INTEGER *sno = NULL;
+
+					CMS_RecipientEncryptedKey *rek = sk_CMS_RecipientEncryptedKey_value(reks, j);
+					CMS_RecipientEncryptedKey_get0_id(rek, &keyid, NULL, NULL, &issuer, &sno);
+
+					bio_print_recipient(text, i, algo_buf, keyid, issuer, sno);
+				}
+			} break;
+			default:
+				g_warning("Unknown recipient information for recipient %d", i);
+				break;
+		}
+	}
+
+	ret = bio_mem_unwrap(text);
+	if (!ret) {
+		g_set_error_literal(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_GET_SIGNER,
+				"Failed to obtain recipient infos: ");
+	}
 
 	return ret;
 }
@@ -744,20 +914,6 @@ static gchar* get_cert_time(const ASN1_TIME *time)
 	return ret;
 }
 
-static gchar* bio_mem_unwrap(BIO *mem)
-{
-	long size;
-	gchar *data, *ret;
-
-	g_return_val_if_fail(mem != NULL, NULL);
-
-	size = BIO_get_mem_data(mem, &data);
-	ret = g_strndup(data, size);
-	BIO_free(mem);
-
-	return ret;
-}
-
 gchar* format_cert_chain(STACK_OF(X509) *verified_chain)
 {
 	BIO *text = NULL;
@@ -903,6 +1059,8 @@ gboolean cms_get_cert_chain(CMS_ContentInfo *cms, X509_STORE *store, STACK_OF(X5
 out:
 	if (cert_ctx)
 		X509_STORE_CTX_free(cert_ctx);
+	if (intercerts)
+		sk_X509_pop_free(intercerts, X509_free);
 	if (signers)
 		sk_X509_free(signers);
 
@@ -957,21 +1115,27 @@ static void debug_cms_ci(CMS_ContentInfo *cms)
 
 gboolean cms_is_detached(GBytes *sig, gboolean *detached, GError **error)
 {
-	CMS_ContentInfo *cms = NULL;
-	BIO *insig = BIO_new_mem_buf((void *)g_bytes_get_data(sig, NULL),
-			g_bytes_get_size(sig));
+	g_autoptr(CMS_ContentInfo) cms = NULL;
+	BIO *insig = NULL;
 	gboolean res = FALSE;
 
 	g_return_val_if_fail(sig != NULL, FALSE);
 	g_return_val_if_fail(detached != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
+	g_assert(g_bytes_get_size(sig) > 0);
+
+	insig = BIO_new_mem_buf((void *)g_bytes_get_data(sig, NULL),
+			g_bytes_get_size(sig));
+	if (!insig)
+		g_error("BIO_new_mem_buf() failed");
+
 	if (!(cms = d2i_CMS_bio(insig, NULL))) {
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_PARSE,
-				"failed to parse signature");
+				"Signature data is no valid CMS");
 		goto out;
 	}
 
@@ -980,14 +1144,36 @@ gboolean cms_is_detached(GBytes *sig, gboolean *detached, GError **error)
 	res = TRUE;
 
 out:
-	if (cms)
-		CMS_ContentInfo_free(cms);
+	BIO_free(insig);
+	return res;
+}
+
+gboolean cms_is_envelopeddata(GBytes *cms_data)
+{
+	g_autoptr(CMS_ContentInfo) cms = NULL;
+	BIO *insig = NULL;
+	gboolean res = FALSE;
+
+	g_return_val_if_fail(cms_data != NULL, FALSE);
+
+	insig = BIO_new_mem_buf((void *)g_bytes_get_data(cms_data, NULL),
+			g_bytes_get_size(cms_data));
+	if (!insig)
+		g_error("BIO_new_mem_buf() failed");
+
+	if (!(cms = d2i_CMS_bio(insig, NULL)))
+		goto out;
+
+	res = (OBJ_obj2nid(CMS_get0_type(cms)) == NID_pkcs7_enveloped);
+
+out:
+	BIO_free(insig);
 	return res;
 }
 
 gboolean cms_get_unverified_manifest(GBytes *sig, GBytes **manifest, GError **error)
 {
-	CMS_ContentInfo *cms = NULL;
+	g_autoptr(CMS_ContentInfo) cms = NULL;
 	BIO *insig = BIO_new_mem_buf((void *)g_bytes_get_data(sig, NULL),
 			g_bytes_get_size(sig));
 	ASN1_OCTET_STRING **content = NULL;
@@ -1047,15 +1233,14 @@ gboolean cms_get_unverified_manifest(GBytes *sig, GBytes **manifest, GError **er
 	res = TRUE;
 
 out:
-	if (cms)
-		CMS_ContentInfo_free(cms);
+	BIO_free(insig);
 	return res;
 }
 
 gboolean cms_verify_bytes(GBytes *content, GBytes *sig, X509_STORE *store, CMS_ContentInfo **cms, GBytes **manifest, GError **error)
 {
 	GError *ierror = NULL;
-	CMS_ContentInfo *icms = NULL;
+	g_autoptr(CMS_ContentInfo) icms = NULL;
 	BIO *incontent = NULL;
 	BIO *insig = BIO_new_mem_buf((void *)g_bytes_get_data(sig, NULL),
 			g_bytes_get_size(sig));
@@ -1163,15 +1348,11 @@ gboolean cms_verify_bytes(GBytes *content, GBytes *sig, X509_STORE *store, CMS_C
 	else
 		verified = CMS_verify(icms, NULL, store, NULL, outcontent, CMS_BINARY);
 	if (!verified) {
-		unsigned long err;
-		const gchar *data;
-		int flags;
-		err = ERR_get_error_line_data(NULL, NULL, &data, &flags);
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_INVALID,
-				"signature verification failed: %s", (flags & ERR_TXT_STRING) ? data : ERR_error_string(err, NULL));
+				"signature verification failed: %s", get_openssl_err_string());
 		goto out;
 	}
 
@@ -1196,7 +1377,7 @@ gboolean cms_verify_bytes(GBytes *content, GBytes *sig, X509_STORE *store, CMS_C
 	}
 
 	if (cms)
-		*cms = icms;
+		*cms = g_steal_pointer(&icms);
 
 	res = TRUE;
 out:
@@ -1204,8 +1385,6 @@ out:
 	BIO_free_all(incontent);
 	BIO_free_all(insig);
 	BIO_free_all(outcontent);
-	if (!cms)
-		CMS_ContentInfo_free(icms);
 	r_context_end_step("cms_verify", res);
 	return res;
 }
@@ -1339,3 +1518,157 @@ gboolean cms_verify_sig(GBytes *sig, X509_STORE *store, CMS_ContentInfo **cms, G
 out:
 	return res;
 }
+
+GBytes *cms_encrypt(GBytes *content, gchar **recipients, GError **error)
+{
+	GError *ierror = NULL;
+	BIO *incontent = NULL;
+	BIO *outsig = BIO_new(BIO_s_mem());
+	STACK_OF(X509) *recipcerts = NULL;
+	g_autoptr(CMS_ContentInfo) cms = NULL;
+	GBytes *res = NULL;
+
+	g_return_val_if_fail(content, NULL);
+	g_return_val_if_fail(recipients, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	incontent = BIO_new_mem_buf((void *)g_bytes_get_data(content, NULL),
+			g_bytes_get_size(content));
+	if (!incontent)
+		g_error("BIO_new_mem_buf() failed");
+
+	recipcerts = sk_X509_new_null();
+
+	/* load all recipient certificates from all provided PEM files */
+	for (gchar **recipcertpath = recipients; recipcertpath && *recipcertpath != NULL; recipcertpath++) {
+		STACK_OF(X509) *filecerts = load_certs_from_file(*recipcertpath, &ierror);
+		if (filecerts == NULL) {
+			g_propagate_error(error, ierror);
+			goto out;
+		}
+
+		/* add all recipient certs from recent file */
+		for (gint i = 0; i < sk_X509_num(filecerts); i++) {
+			sk_X509_push(recipcerts, sk_X509_value(filecerts, i));
+		}
+
+		sk_X509_free(filecerts);
+	}
+
+	cms = CMS_encrypt(recipcerts, incontent, EVP_aes_256_cbc(), CMS_BINARY);
+	if (cms == NULL) {
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_INVALID,
+				"Failed to encrypt: %s", get_openssl_err_string());
+		goto out;
+	}
+	if (!i2d_CMS_bio(outsig, cms)) {
+		g_set_error_literal(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_SERIALIZE_SIG,
+				"Failed to serialize signature");
+		goto out;
+	}
+
+	res = bytes_from_bio(outsig);
+	if (!res) {
+		g_set_error_literal(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_UNKNOWN,
+				"Read zero bytes");
+		goto out;
+	}
+
+	g_message("Encrypted for %d recipient%s", sk_X509_num(recipcerts), sk_X509_num(recipcerts) > 1 ? "s" : "");
+
+out:
+	ERR_print_errors_fp(stdout);
+	sk_X509_pop_free(recipcerts, X509_free);
+	BIO_free_all(incontent);
+	BIO_free_all(outsig);
+	return res;
+}
+
+GBytes *cms_decrypt(GBytes *content, const gchar *certfile, const gchar *keyfile, GError **error)
+{
+	GError *ierror = NULL;
+	g_autoptr(CMS_ContentInfo) icms = NULL;
+	g_autoptr(X509) decrypt_cert = NULL;
+	g_autoptr(EVP_PKEY) privkey = NULL;
+	BIO *outdecrypt = BIO_new(BIO_s_mem());
+	BIO *inenc = NULL;
+	GBytes *res = NULL;
+
+	g_return_val_if_fail(content != NULL, NULL);
+	g_return_val_if_fail(keyfile != NULL, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	inenc = BIO_new_mem_buf((void *)g_bytes_get_data(content, NULL),
+			g_bytes_get_size(content));
+	if (!inenc)
+		g_error("BIO_new_mem_buf() failed");
+
+	if (certfile) {
+		decrypt_cert = load_cert(certfile, &ierror);
+		if (decrypt_cert == NULL) {
+			g_propagate_error(error, ierror);
+			res = NULL;
+			goto out;
+		}
+	}
+
+	privkey = load_key(keyfile, &ierror);
+	if (privkey == NULL) {
+		g_propagate_error(error, ierror);
+		res = NULL;
+		goto out;
+	}
+
+	g_message("Decrypting signature...");
+
+	if (!(icms = d2i_CMS_bio(inenc, NULL))) {
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_PARSE,
+				"Failed to parse CMS");
+		res = NULL;
+		goto out;
+	}
+
+	/* assert we received envelopedData */
+	if (OBJ_obj2nid(CMS_get0_type(icms)) != NID_pkcs7_enveloped) {
+		g_set_error(error, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID, "Expected CMS of type '%s' but got '%s'", OBJ_nid2sn(NID_pkcs7_enveloped), OBJ_nid2sn(OBJ_obj2nid(CMS_get0_type(icms))));
+		res = NULL;
+		goto out;
+	}
+
+	if (!CMS_decrypt(icms, privkey, decrypt_cert, NULL, outdecrypt, 0)) {
+		res = NULL;
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_INVALID,
+				"Failed to decrypt CMS EnvelopedData: %s", get_openssl_err_string());
+		goto out;
+	}
+
+	res = bytes_from_bio(outdecrypt);
+	if (!res) {
+		g_set_error_literal(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_UNKNOWN,
+				"Read zero bytes");
+		goto out;
+	}
+out:
+	ERR_print_errors_fp(stdout);
+	BIO_free_all(inenc);
+	BIO_free_all(outdecrypt);
+	return res;
+}
