@@ -17,7 +17,7 @@
 #include "mbr.h"
 #include "gpt.h"
 #include "utils.h"
-
+#include "hash_index.h"
 
 #define R_SLOT_HOOK_PRE_INSTALL "slot-pre-install"
 #define R_SLOT_HOOK_POST_INSTALL "slot-post-install"
@@ -34,7 +34,6 @@ GQuark r_update_error_quark(void)
 static GUnixOutputStream* open_slot_device(RaucSlot *slot, int *fd, GError **error)
 {
 	GUnixOutputStream *outstream = NULL;
-	GError *ierror = NULL;
 	int fd_out;
 
 	g_return_val_if_fail(slot, NULL);
@@ -50,8 +49,8 @@ static GUnixOutputStream* open_slot_device(RaucSlot *slot, int *fd, GError **err
 
 	outstream = G_UNIX_OUTPUT_STREAM(g_unix_output_stream_new(fd_out, TRUE));
 	if (outstream == NULL) {
-		g_propagate_prefixed_error(error, ierror,
-				"Failed to open file for writing: ");
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Failed to create stream for output device %s", slot->device);
 		return NULL;
 	}
 
@@ -181,13 +180,14 @@ static gboolean ubifs_ioctl(RaucImage *image, int fd, GError **error)
 	return TRUE;
 }
 
-static gboolean copy_raw_image(RaucImage *image, GUnixOutputStream *outstream, GError **error)
+static gboolean copy_raw_image(RaucImage *image, GUnixOutputStream *outstream, gsize len_header_last, GError **error)
 {
 	GError *ierror = NULL;
 	gssize size;
 	goffset seeksize;
 	g_autoptr(GFile) srcimagefile = g_file_new_for_path(image->filename);
 	int out_fd = g_unix_output_stream_get_fd(outstream);
+	g_autofree void *header = NULL;
 
 	g_autoptr(GInputStream) instream = G_INPUT_STREAM(g_file_read(srcimagefile, NULL, &ierror));
 	if (instream == NULL) {
@@ -196,6 +196,29 @@ static gboolean copy_raw_image(RaucImage *image, GUnixOutputStream *outstream, G
 		return FALSE;
 	}
 
+	if (len_header_last) {
+		gsize sector_size = (gsize) get_sectorsize(out_fd);
+
+		if (len_header_last != sector_size) {
+			g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+					"Specified header length (%"G_GSIZE_FORMAT ") does not match sector size (%"G_GSIZE_FORMAT ")", len_header_last, sector_size);
+			return FALSE;
+		}
+
+		header = g_malloc(len_header_last);
+
+		if (!g_input_stream_read_all(instream, header, len_header_last, &len_header_last, NULL, &ierror)) {
+			g_propagate_prefixed_error(error, ierror,
+					"Failed to read header: ");
+			return FALSE;
+		}
+
+		if (lseek(out_fd, len_header_last, SEEK_CUR) == -1) {
+			g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED, "Failed to skip header: %s", strerror(errno));
+			return FALSE;
+		}
+	}
+
 	size = g_output_stream_splice(G_OUTPUT_STREAM(outstream), instream,
 			G_OUTPUT_STREAM_SPLICE_NONE,
 			NULL,
@@ -214,6 +237,26 @@ static gboolean copy_raw_image(RaucImage *image, GUnixOutputStream *outstream, G
 		return FALSE;
 	}
 
+	if (len_header_last) {
+		gsize bytes;
+
+		if (fsync(out_fd) == -1) {
+			g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED, "Syncing content to disk failed: %s", strerror(errno));
+			return FALSE;
+		}
+
+		if (lseek(out_fd, -seeksize, SEEK_CUR) == -1) {
+			g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED, "Failed to rewind output stream: %s", strerror(errno));
+			return FALSE;
+		}
+
+		if (!g_output_stream_write_all(G_OUTPUT_STREAM(outstream), header, len_header_last, &bytes, NULL, &ierror)) {
+			g_propagate_prefixed_error(error, ierror,
+					"Failed to write header: ");
+			return FALSE;
+		}
+	}
+
 	if (!g_input_stream_close(instream, NULL, &ierror)) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -228,15 +271,81 @@ static gboolean copy_raw_image(RaucImage *image, GUnixOutputStream *outstream, G
 	return TRUE;
 }
 
+static gboolean write_boot_switch_partition(RaucImage *image, const gchar *device,
+		const struct boot_switch_partition *dest_partition,
+		gsize len_header_last,
+		GError **error)
+
+{
+	GError *ierror = NULL;
+	gboolean res = FALSE;
+	int out_fd = -1;
+	g_autoptr(GUnixOutputStream) outstream = NULL;
+
+	g_return_val_if_fail(image, FALSE);
+	g_return_val_if_fail(device, FALSE);
+	g_return_val_if_fail(dest_partition, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	out_fd = open(device, O_WRONLY);
+	if (out_fd == -1) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Opening output device failed: %s",
+				strerror(errno));
+		res = FALSE;
+		goto out;
+	}
+
+	if (lseek(out_fd, dest_partition->start, SEEK_SET) !=
+	    (off_t)dest_partition->start) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Failed to set file to position %"G_GUINT64_FORMAT,
+				dest_partition->start);
+		res = FALSE;
+		goto out;
+	}
+
+	outstream = G_UNIX_OUTPUT_STREAM(g_unix_output_stream_new(out_fd, FALSE));
+	if (outstream == NULL) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Failed to create output stream");
+		res = FALSE;
+		goto out;
+	}
+
+	res = copy_raw_image(image, outstream, len_header_last, &ierror);
+	if (!res) {
+		g_propagate_error(error, ierror);
+		goto out;
+	}
+
+	res = g_output_stream_close(G_OUTPUT_STREAM(outstream), NULL, &ierror);
+	if (!res) {
+		g_propagate_error(error, ierror);
+		goto out;
+	}
+
+out:
+	if (out_fd >= 0)
+		close(out_fd);
+
+	return res;
+}
+
 static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const gchar *seed, const gchar *store, const gchar *tmpdir, GError **error)
 {
 	g_autoptr(GSubprocessLauncher) launcher = NULL;
 	g_autoptr(GSubprocess) sproc = NULL;
+	g_auto(GStrv) casync_argvp = NULL;
 	GError *ierror = NULL;
 	gboolean res = FALSE;
 	g_autoptr(GPtrArray) args = g_ptr_array_new_full(5, g_free);
 
-	g_ptr_array_add(args, g_strdup("casync"));
+	if (r_context()->config->use_desync)
+		g_ptr_array_add(args, g_strdup("desync"));
+	else
+		g_ptr_array_add(args, g_strdup("casync"));
+
 	g_ptr_array_add(args, g_strdup("extract"));
 	if (seed) {
 		g_ptr_array_add(args, g_strdup("--seed"));
@@ -246,7 +355,25 @@ static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const
 		g_ptr_array_add(args, g_strdup("--store"));
 		g_ptr_array_add(args, g_strdup(store));
 	}
-	g_ptr_array_add(args, g_strdup("--seed-output=no"));
+	/* Desync doesn't have the --seed-output option */
+	if (!r_context()->config->use_desync)
+		g_ptr_array_add(args, g_strdup("--seed-output=no"));
+
+	if (r_context()->config->casync_install_args != NULL) {
+		gboolean parse_res = FALSE;
+		parse_res = g_shell_parse_argv(r_context()->config->casync_install_args, NULL, &casync_argvp, &ierror);
+		if (!parse_res) {
+			res = parse_res;
+			g_propagate_prefixed_error(
+					error,
+					ierror,
+					"Failed to parse casync extra args: ");
+			goto out;
+		}
+		for (gchar **casync_args = casync_argvp; *casync_args != NULL; casync_args++) {
+			g_ptr_array_add(args, g_strdup(*casync_args));
+		}
+	}
 	g_ptr_array_add(args, g_strdup(image->filename));
 	g_ptr_array_add(args, g_strdup(out_fd >= 0 ? "-" : dest));
 	g_ptr_array_add(args, NULL);
@@ -257,6 +384,10 @@ static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const
 	if (tmpdir)
 		g_subprocess_launcher_setenv(launcher, "TMPDIR", tmpdir, TRUE);
 
+	/* Enable Desync parsable progress updates */
+	if (r_context()->config->use_desync)
+		g_subprocess_launcher_setenv(launcher, "DESYNC_ENABLE_PARSABLE_PROGRESS", "1", TRUE);
+
 	sproc = r_subprocess_launcher_spawnv(launcher, args, &ierror);
 	if (sproc == NULL) {
 		g_propagate_prefixed_error(
@@ -286,6 +417,11 @@ static RaucSlot *get_active_slot_class_member(gchar *slotclass)
 
 	g_return_val_if_fail(slotclass, NULL);
 
+	if (!r_context()->config->slots) {
+		/* when no slots are configured, there can be not active slot */
+		return NULL;
+	}
+
 	g_hash_table_iter_init(&iter, r_context()->config->slots);
 	while (g_hash_table_iter_next(&iter, NULL, (gpointer *)&iterslot)) {
 		if (iterslot->state == ST_INACTIVE)
@@ -309,10 +445,19 @@ static gboolean casync_extract_image(RaucImage *image, gchar *dest, int out_fd,
 	gchar *tmpdir = NULL;
 	gboolean seed_mounted = FALSE;
 
+	g_assert_nonnull(r_context()->install_info);
+	g_assert_nonnull(r_context()->install_info->mounted_bundle);
+	g_assert_nonnull(r_context()->install_info->mounted_bundle->storepath);
+
+	if (r_context()->config->use_desync) {
+		/* TODO: do something clever to locate and/or generate the seed index file */
+		goto extract;
+	}
+
 	/* Prepare Seed */
 	seedslot = get_active_slot_class_member(image->slotclass);
 	if (!seedslot) {
-		g_warning("No seed slot available for %s", image->slotclass);
+		g_message("No casync seed slot available for %s", image->slotclass);
 		goto extract;
 	}
 
@@ -322,7 +467,7 @@ static gboolean casync_extract_image(RaucImage *image, gchar *dest, int out_fd,
 		 * rootfs slot seed is inaproppriate as it contains virtual
 		 * file systems, additional mounts, etc. */
 		if (!seedslot->mount_point) {
-			g_debug("Mounting %s to use as seed", seedslot->device);
+			g_debug("Mounting %s to use as casync seed", seedslot->device);
 			res = r_mount_slot(seedslot, &ierror);
 			if (!res) {
 				g_warning("Failed mounting for seeding: %s", ierror->message);
@@ -340,8 +485,10 @@ static gboolean casync_extract_image(RaucImage *image, gchar *dest, int out_fd,
 		/* For the moment do not utilize UBI volumes as seed because they are
 		 * character devices - additional logic is needed to (temporarily) map
 		 * them to UBIBLOCK devices which are suitable for that purpose */
-		if (g_stat(seedslot->device, &seedstat) < 0 || S_ISCHR(seedstat.st_mode))
+		if (g_stat(seedslot->device, &seedstat) < 0 || S_ISCHR(seedstat.st_mode)) {
+			g_message("Cannot use %s as seed device (non-existing or char device)", seedslot->device);
 			goto extract;
+		}
 
 		g_debug("Adding as casync blob seed: %s", seedslot->device);
 		seed = g_strdup(seedslot->device);
@@ -350,31 +497,39 @@ static gboolean casync_extract_image(RaucImage *image, gchar *dest, int out_fd,
 extract:
 	/* Set store */
 	store = r_context()->install_info->mounted_bundle->storepath;
-	g_debug("Using store path: '%s'", store);
+	g_debug("Using casync store path: '%s'", store);
 
 	/* Set temporary directory */
 	tmpdir = r_context()->config->tmp_path;
 	if (tmpdir)
-		g_debug("Using tmp path: '%s'", tmpdir);
+		g_debug("Using casync tmp path: '%s'", tmpdir);
 
 	/* Call casync to extract */
 	res = casync_extract(image, dest, out_fd, seed, store, tmpdir, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
-		goto out;
-	}
-
-	/* Cleanup seed */
-	if (seed_mounted) {
-		res = r_umount_slot(seedslot, &ierror);
-		if (!res) {
-			g_propagate_prefixed_error(error, ierror, "Failed unmounting seed slot: ");
-			goto out;
-		}
+		goto unmount_out;
 	}
 
 	res = TRUE;
-out:
+
+unmount_out:
+	/* Cleanup seed */
+	if (seed_mounted) {
+		g_message("Unmounting seed slot %s", seedslot->device);
+		ierror = NULL; /* any previous error was propagated already */
+		if (!r_umount_slot(seedslot, &ierror)) {
+			res = FALSE;
+			if (error && *error) {
+				/* the previous error is more relevant here */
+				g_warning("Ignoring umount error after previous error: %s", ierror->message);
+				g_clear_error(&ierror);
+			} else {
+				g_propagate_error(error, ierror);
+			}
+		}
+	}
+
 	return res;
 }
 
@@ -395,7 +550,7 @@ static gboolean copy_raw_image_to_dev(RaucImage *image, RaucSlot *slot, GError *
 
 	/* copy */
 	g_message("writing data to device %s", slot->device);
-	res = copy_raw_image(image, outstream, &ierror);
+	res = copy_raw_image(image, outstream, 0, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
 		goto out;
@@ -411,31 +566,264 @@ out:
 	return res;
 }
 
-static gboolean write_image_to_dev(RaucImage *image, RaucSlot *slot, GError **error)
+static gboolean copy_block_hash_index_image_to_dev(RaucImage *image, RaucSlot *slot, GError **error)
 {
 	GError *ierror = NULL;
 	gboolean res = FALSE;
+	g_autoptr(RaucHashIndex) tmp = NULL;
+	g_autoptr(GPtrArray) sources = NULL;
+	const RaucSlot *seedslot = NULL;
+	const guint8(*chunk_hashes)[32];
+	guint32 chunk_count;
+	g_autofree RaucHashIndexChunk *chunk = NULL;
+	off_t offset = 0;
+	int target_fd = -1;
+	g_autoptr(RaucStats) zero_stats = NULL;
+
+	g_return_val_if_fail(image, FALSE);
+	g_return_val_if_fail(slot, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	zero_stats = r_stats_new("zero chunk");
+
+	sources = g_ptr_array_new_with_free_func((GDestroyNotify)r_hash_index_free);
+
+	/* If we have an index for the target slot, use it, otherwise generate and append for upper range. */
+	/* Compared to open_slot_device, we need O_RDWR and seeking. */
+	tmp = r_hash_index_open_slot("target_slot", slot, O_RDWR | O_EXCL, &ierror);
+	if (!tmp) {
+		g_propagate_prefixed_error(error, ierror, "failed to open target slot hash index for %s: ", slot->name);
+		res = FALSE;
+		goto out;
+	}
+	g_ptr_array_add(sources, g_steal_pointer(&tmp));
+
+	/* Open and append seed slot. */
+	seedslot = get_active_slot_class_member(image->slotclass);
+	if (seedslot) {
+		tmp = r_hash_index_open_slot("active_slot", seedslot, O_RDONLY, &ierror);
+		if (!tmp) {
+			g_propagate_prefixed_error(error, ierror, "failed to open active slot hash index for %s: ", seedslot->name);
+			res = FALSE;
+			goto out;
+		}
+		g_ptr_array_add(sources, g_steal_pointer(&tmp));
+	} else {
+		g_message("No active slot available to use as seed for %s", image->slotclass);
+	}
+
+	/* Open and append source image. */
+	tmp = r_hash_index_open_image("source_image", image, &ierror);
+	if (!tmp) {
+		g_propagate_prefixed_error(error, ierror, "failed to open source image hash index for %s: ", image->filename);
+		res = FALSE;
+		goto out;
+	}
+	/* The bundle data is read-only and authenticated. */
+	tmp->skip_hash_check = TRUE;
+	g_ptr_array_add(sources, g_steal_pointer(&tmp));
+
+	/* Open source index and target fd for lower range (reuse written chunks). */
+	tmp = r_hash_index_reuse("target_slot_written",
+			g_ptr_array_index(sources, sources->len - 1),
+			dup(((RaucHashIndex*)g_ptr_array_index(sources, 0))->data_fd),
+			&ierror);
+	if (!tmp) {
+		g_propagate_prefixed_error(error, ierror, "failed to reuse source hash index for target slot: ");
+		res = FALSE;
+		goto out;
+	}
+	/* Nothing is valid yet. */
+	tmp->invalid_from = 0;
+	/* This is the data we've just written. */
+	tmp->skip_hash_check = TRUE;
+	/* Insert the growing region in the target slot first. */
+	g_ptr_array_insert(sources, 0, g_steal_pointer(&tmp));
+
+	/* Now we have a GPtrArray of RaucHashIndices:
+	 *
+	 * 0: target slot with source index (for reuse of written chunks)
+	 * 1: target slot with corresponding old index
+	 * 2: active slot with corresponding index (optional)
+	 * len-1: source image with corresponding index
+	 */
+	g_assert(sources->len <= 4);
+
+	{
+		const RaucHashIndex *target = g_ptr_array_index(sources, 0);
+		const RaucHashIndex *source = g_ptr_array_index(sources, sources->len-1);
+		target_fd = target->data_fd;
+		chunk_hashes = g_bytes_get_data(source->hashes, NULL);
+		chunk_count = source->count;
+	}
+
+	/* Ensure we start writing from the beginning */
+	offset = 0;
+	if (lseek(target_fd, offset, SEEK_SET) != offset) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED, "Failed to seek to start of target slot: %s", g_strerror(errno));
+		res = FALSE;
+		goto out;
+	}
+
+	/* Temporary data storage */
+	chunk = g_new0(RaucHashIndexChunk, 1);
+
+	/* Iterate over chunks in source image */
+	for (guint32 c = 0; c < chunk_count; c++) {
+		gboolean found = FALSE;
+
+		if (memcmp(chunk_hashes[c], R_HASH_INDEX_ZERO_CHUNK, 32) == 0) {
+			/* Generate zero chunk */
+			memset(chunk->data, 0, sizeof(chunk->data));
+			found = TRUE;
+			r_stats_add(zero_stats, 1);
+		} else {
+			/* Iterate over indices and call get chunk */
+			for (guint s = 0; s < sources->len; s++) {
+				const RaucHashIndex *source = g_ptr_array_index(sources, s);
+				if (r_hash_index_get_chunk(source, chunk_hashes[c], chunk, &ierror)) {
+					//g_autofree gchar *hash = r_hex_encode(chunk_hashes[c], sizeof(chunk_hashes[c]));
+					//g_debug("found chunk %"G_GUINT32_FORMAT" [%s] in index %u [%s]", c, hash, s, source->label);
+					found = TRUE;
+					break;
+				} else {
+					//g_autofree gchar *hash = r_hex_encode(chunk_hashes[c], sizeof(chunk_hashes[c]));
+					//g_debug("no chunk %"G_GUINT32_FORMAT" [%s] in index %u [%s]: %s", c, hash, s, source->label, ierror->message);
+					g_clear_error(&ierror);
+				}
+			}
+		}
+
+		if (!found) {
+			g_autofree gchar *hash = r_hex_encode(chunk_hashes[c], sizeof(chunk_hashes[c]));
+			g_set_error(error,
+					R_HASH_INDEX_ERROR,
+					R_HASH_INDEX_ERROR_NOT_FOUND,
+					"no chunk with required hash [%s] found", hash);
+			res = FALSE;
+			goto out;
+		}
+
+		/* Write chunk to target
+		 *
+		 * We could potentially avoid a chunk write if  r_hash_index_get_chunk
+		 * would report where the chunk was found. If it was found on the target
+		 * in the correct location, we could skip the write.
+		 */
+		offset = (off_t)c * sizeof(chunk->data);
+		if (!r_pwrite_lazy(target_fd, chunk->data, sizeof(chunk->data), offset, &ierror)) {
+			g_propagate_error(error, ierror);
+			res = FALSE;
+			goto out;
+		}
+
+		/* Update limits */
+		{
+			RaucHashIndex *target_written = g_ptr_array_index(sources, 0);
+			RaucHashIndex *target_old = g_ptr_array_index(sources, 1);
+			target_written->invalid_from = c+1;
+			target_old->invalid_below = c;
+		}
+	}
+
+	/* Seek after the written data so this behaves similar to the simpler write helpers */
+	offset = (off_t)chunk_count * sizeof(chunk->data);
+	if (lseek(target_fd, offset, SEEK_SET) != offset) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED, "Failed to seek to end of image: %s", g_strerror(errno));
+		res = FALSE;
+		goto out;
+	}
+
+
+	/* Flush to block device before closing to assure content is written to disk */
+	if (fsync(target_fd) == -1) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED, "Syncing content to slot failed: %s", strerror(errno));
+		res = FALSE;
+		goto out;
+	}
+
+	/* Write new index to slot data dir. */
+	{
+		const RaucHashIndex *source = g_ptr_array_index(sources, sources->len-1);
+		if (!r_hash_index_export_slot(source, slot, &image->checksum, &ierror)) {
+			g_warning("Continuing after failure to write new hash index: %s", ierror->message);
+		}
+	}
+
+	r_stats_show(zero_stats, "access stats for");
+	for (guint s = 0; s < sources->len; s++) {
+		const RaucHashIndex *source = g_ptr_array_index(sources, s);
+		r_stats_show(source->match_stats, "access stats for");
+	}
+
+	res = TRUE;
+
+out:
+	/* We let the hash index close the file and use dup for the target slot, to simplify cleanup */
+	return res;
+}
+
+static gboolean copy_adaptive_image_to_dev(RaucImage *image, RaucSlot *slot, GError **error)
+{
+	GError *ierror = NULL;
+	g_autofree gchar* temp_string = NULL;
+
+	g_return_val_if_fail(image, FALSE);
+	g_return_val_if_fail(slot, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	if (g_strv_contains((const gchar * const*)image->adaptive, "block-hash-index")) {
+		if (!copy_block_hash_index_image_to_dev(image, slot, &ierror)) {
+			g_propagate_error(error, ierror);
+			return FALSE;
+		}
+		return TRUE;
+	}
+
+	temp_string = g_strjoinv(" ", (gchar**) image->adaptive);
+	g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_UNSUPPORTED_ADAPTIVE_MODE,
+			"No compatible adaptive mode found in '%s'", temp_string);
+	return FALSE;
+}
+
+static gboolean write_image_to_dev(RaucImage *image, RaucSlot *slot, GError **error)
+{
+	GError *ierror = NULL;
 
 	/* Handle casync index file */
 	if (g_str_has_suffix(image->filename, ".caibx")) {
 		g_message("Extracting %s to %s", image->filename, slot->device);
 
 		/* Extract caibx to device */
-		res = casync_extract_image(image, slot->device, -1, &ierror);
-		if (!res) {
+		if (!casync_extract_image(image, slot->device, -1, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
-	} else {
-		res = copy_raw_image_to_dev(image, slot, &ierror);
-		if (!res) {
-			g_propagate_error(error, ierror);
-			goto out;
+		return TRUE;
+	}
+
+	/* Try adaptive mode */
+	if (image->adaptive && slot->data_directory) {
+		if (!copy_adaptive_image_to_dev(image, slot, &ierror)) {
+			if (g_error_matches(ierror, R_UPDATE_ERROR, R_UPDATE_ERROR_UNSUPPORTED_ADAPTIVE_MODE)) {
+				g_info("%s", ierror->message);
+			} else {
+				g_warning("Continuing after adaptive mode error: %s", ierror->message);
+			}
+			g_clear_error(&ierror);
+			/* Continue with full copy */
+		} else {
+			return TRUE;
 		}
 	}
 
-out:
-	return res;
+	/* Finally, try a raw copy */
+	if (!copy_raw_image_to_dev(image, slot, &ierror)) {
+		g_propagate_error(error, ierror);
+		return FALSE;
+	}
+
+	return TRUE;
 }
 
 static gboolean ubifs_format_slot(RaucSlot *dest_slot, GError **error)
@@ -578,7 +966,6 @@ static gchar* vfat_label_generator(const gchar *name)
 			*c = g_ascii_toupper(*c);
 		/* replace invalid chars */
 		for (size_t i = 0; i < strlen(invalid_chars); i++) {
-			g_message("r: %c", invalid_chars[i]);
 			if (*c == invalid_chars[i]) {
 				*c = '_';
 				break;
@@ -988,7 +1375,7 @@ static gboolean img_to_ubivol_handler(RaucImage *image, RaucSlot *dest_slot, con
 		}
 	} else {
 		/* copy */
-		res = copy_raw_image(image, outstream, &ierror);
+		res = copy_raw_image(image, outstream, 0, &ierror);
 		if (!res) {
 			g_propagate_error(error, ierror);
 			goto out;
@@ -1057,7 +1444,7 @@ static gboolean img_to_ubifs_handler(RaucImage *image, RaucSlot *dest_slot, cons
 		}
 	} else {
 		/* copy */
-		res = copy_raw_image(image, outstream, &ierror);
+		res = copy_raw_image(image, outstream, 0, &ierror);
 		if (!res) {
 			g_propagate_error(error, ierror);
 			goto out;
@@ -1487,8 +1874,7 @@ out:
 static gboolean img_to_boot_mbr_switch_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	gboolean res = FALSE;
-	int out_fd = -1, inactive_half;
-	g_autoptr(GUnixOutputStream) outstream = NULL;
+	int inactive_half;
 	GError *ierror = NULL;
 	struct boot_switch_partition dest_partition;
 	g_autoptr(GHashTable) vars = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
@@ -1544,47 +1930,12 @@ static gboolean img_to_boot_mbr_switch_handler(RaucImage *image, RaucSlot *dest_
 		goto out;
 	}
 
-	g_message("Opening inactive (%s) half of boot partition region on %s", inactive_half == 0 ? "first" : "second",
-			dest_slot->device);
+	g_message("Write image to inactive (%s) half of boot partition region on %s", inactive_half == 0 ? "first" : "second", dest_slot->device);
 
-	out_fd = open(dest_slot->device, O_WRONLY);
-	if (out_fd == -1) {
-		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
-				"Opening output device failed: %s",
-				strerror(errno));
-		res = FALSE;
-		goto out;
-	}
-
-	if (lseek(out_fd, dest_partition.start, SEEK_SET) !=
-	    (off_t)dest_partition.start) {
-		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
-				"Failed to set file to position %"G_GUINT64_FORMAT,
-				dest_partition.start);
-		res = FALSE;
-		goto out;
-	}
-
-	outstream = G_UNIX_OUTPUT_STREAM(g_unix_output_stream_new(out_fd, FALSE));
-	if (outstream == NULL) {
+	res = write_boot_switch_partition(image, dest_slot->device, &dest_partition, 0, &ierror);
+	if (!res) {
 		g_propagate_prefixed_error(error, ierror,
-				"Failed to open file for writing: ");
-		res = FALSE;
-		goto out;
-	}
-
-	g_message("Copying image to inactive (%s) half of boot partition region on %s",
-			inactive_half == 0 ? "first" : "second", dest_slot->device);
-
-	res = copy_raw_image(image, outstream, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
-		goto out;
-	}
-
-	res = g_output_stream_close(G_OUTPUT_STREAM(outstream), NULL, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
+				"Failed to write inactive region: ");
 		goto out;
 	}
 
@@ -1607,8 +1958,6 @@ static gboolean img_to_boot_mbr_switch_handler(RaucImage *image, RaucSlot *dest_
 	}
 
 out:
-	if (out_fd >= 0)
-		close(out_fd);
 
 	return res;
 }
@@ -1617,8 +1966,7 @@ G_GNUC_UNUSED
 static gboolean img_to_boot_gpt_switch_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	gboolean res = FALSE;
-	int out_fd = -1, inactive_half;
-	g_autoptr(GUnixOutputStream) outstream = NULL;
+	int inactive_half;
 	GError *ierror = NULL;
 	struct boot_switch_partition dest_partition;
 	g_autoptr(GHashTable) vars = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
@@ -1674,47 +2022,12 @@ static gboolean img_to_boot_gpt_switch_handler(RaucImage *image, RaucSlot *dest_
 		goto out;
 	}
 
-	g_message("Opening inactive (%s) half of boot partition region on %s", inactive_half == 0 ? "first" : "second",
-			dest_slot->device);
+	g_message("Write image to inactive (%s) half of boot partition region on %s", inactive_half == 0 ? "first" : "second", dest_slot->device);
 
-	out_fd = open(dest_slot->device, O_WRONLY);
-	if (out_fd == -1) {
-		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
-				"Opening output device failed: %s",
-				strerror(errno));
-		res = FALSE;
-		goto out;
-	}
-
-	if (lseek(out_fd, dest_partition.start, SEEK_SET) !=
-	    (off_t)dest_partition.start) {
-		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
-				"Failed to set file to position %"G_GUINT64_FORMAT,
-				dest_partition.start);
-		res = FALSE;
-		goto out;
-	}
-
-	outstream = G_UNIX_OUTPUT_STREAM(g_unix_output_stream_new(out_fd, FALSE));
-	if (outstream == NULL) {
+	res = write_boot_switch_partition(image, dest_slot->device, &dest_partition, 0, &ierror);
+	if (!res) {
 		g_propagate_prefixed_error(error, ierror,
-				"Failed to open file for writing: ");
-		res = FALSE;
-		goto out;
-	}
-
-	g_message("Copying image to inactive (%s) half of boot partition region on %s",
-			inactive_half == 0 ? "first" : "second", dest_slot->device);
-
-	res = copy_raw_image(image, outstream, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
-		goto out;
-	}
-
-	res = g_output_stream_close(G_OUTPUT_STREAM(outstream), NULL, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
+				"Failed to write inactive region: ");
 		goto out;
 	}
 
@@ -1736,9 +2049,6 @@ static gboolean img_to_boot_gpt_switch_handler(RaucImage *image, RaucSlot *dest_
 		goto out;
 	}
 out:
-	if (out_fd >= 0)
-		close(out_fd);
-
 	return res;
 }
 
@@ -1840,7 +2150,7 @@ static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot,
 	/* copy */
 	g_message("Copying image to slot device partition %s",
 			part_slot->device);
-	res = copy_raw_image(image, outstream, &ierror);
+	res = copy_raw_image(image, outstream, 0, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
 		goto out;
@@ -1921,6 +2231,179 @@ out:
 }
 #endif
 
+static gboolean check_if_area_is_clear(const gchar *device, guint64 start, gsize size, gboolean *clear, GError **error)
+{
+	gboolean res = FALSE;
+	g_autofree guchar *read_buf = NULL;
+	gint read_size = 512;
+	gint fd;
+
+	g_return_val_if_fail(device, FALSE);
+	g_return_val_if_fail(clear, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	read_buf = g_malloc0(read_size);
+
+	fd = g_open(device, O_RDONLY);
+	if (fd == -1) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Opening device failed: %s",
+				g_strerror(errno));
+		goto out;
+	}
+
+	if (lseek(fd, start, SEEK_SET) != (off_t) start) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Failed to set file to position %"G_GUINT64_FORMAT ": %s",
+				start, g_strerror(errno));
+		goto out;
+	}
+
+	*clear = TRUE;
+
+	while (size && *clear) {
+		gint read_count = 0;
+
+		if (size < (gsize) read_size)
+			read_size = size;
+
+		read_count = read(fd, read_buf, read_size);
+		if (read_count < 0)
+			goto out;
+
+		for (gint i = 0; i < read_count; i++)
+		{
+			if ((read_buf[i] != 0x00) && (read_buf[i] != 0xFF)) {
+				*clear = FALSE;
+				break;
+			}
+		}
+
+		size -= read_count;
+	}
+
+	res = TRUE;
+
+out:
+	if (fd >= 0)
+		g_close(fd, NULL);
+
+	return res;
+}
+
+static gboolean img_to_boot_raw_fallback_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
+{
+	gboolean res = FALSE;
+	GError *ierror = NULL;
+	guint64 half_size;
+	gboolean primary_clear;
+	struct part_desc {
+		const char *name;
+		struct boot_switch_partition partition;
+	} part_desc[2];
+	int first_part_desc_index = 0;
+	g_autoptr(GHashTable) vars = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
+	const gsize header_size = 512;
+
+	g_return_val_if_fail(image, FALSE);
+	g_return_val_if_fail(dest_slot, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	half_size = dest_slot->region_size / 2;
+
+	/* Since down in copy_raw_image() the header size must match the sector size, checking the
+	 * alignment to the header size also implicitly ensures sector alignment.
+	 */
+	if ((dest_slot->region_start % header_size) != 0) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Region start %"G_GUINT64_FORMAT " is not aligned to the header size %"G_GSIZE_FORMAT,
+				dest_slot->region_start, header_size);
+		goto out;
+	}
+
+	if ((half_size % header_size) != 0) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Half region size %"G_GUINT64_FORMAT " is not aligned to the header size %"G_GSIZE_FORMAT,
+				half_size, header_size);
+		goto out;
+	}
+
+	if (half_size < (guint64)image->checksum.size) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Size of image (%"G_GOFFSET_FORMAT ") does not fit to slot size %"G_GUINT64_FORMAT,
+				image->checksum.size, half_size);
+		goto out;
+	}
+
+	g_hash_table_insert(vars, g_strdup("RAUC_BOOT_REGION_START"),
+			g_strdup_printf("%"G_GUINT64_FORMAT, dest_slot->region_start));
+	g_hash_table_insert(vars, g_strdup("RAUC_BOOT_REGION_SIZE"),
+			g_strdup_printf("%"G_GUINT64_FORMAT, dest_slot->region_size));
+
+	/* run slot pre install hook if enabled */
+	if (hook_name && image->hooks.pre_install) {
+		res = run_slot_hook_extra_env(hook_name, R_SLOT_HOOK_PRE_INSTALL, image,
+				dest_slot, vars, &ierror);
+		if (!res) {
+			g_propagate_error(error, ierror);
+			goto out;
+		}
+	}
+
+	part_desc[0].name = "fallback";
+	part_desc[0].partition.start = dest_slot->region_start + half_size;
+	part_desc[0].partition.size = half_size;
+
+	part_desc[1].name = "primary";
+	part_desc[1].partition.start = dest_slot->region_start;
+	part_desc[1].partition.size = half_size;
+
+	/* If the primary partition is not fully programmed, it most likely means that the fallback
+	 * partition was used to boot and is therefore valid. To avoid ending up with two broken partitions,
+	 * upgrade the primary partition first.
+	 */
+	res = check_if_area_is_clear(dest_slot->device, dest_slot->region_start, header_size, &primary_clear, &ierror);
+	if (!res) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Failed to check area at %"G_GUINT64_FORMAT " on %s",
+				dest_slot->region_start, dest_slot->device);
+		goto out;
+	}
+
+	if (primary_clear)
+		first_part_desc_index++;
+
+	for (gint i = 0; i < 2; i++)
+	{
+		struct part_desc *pd = &part_desc[(first_part_desc_index + i) % 2];
+
+		g_message("Updating %s partition at %"G_GUINT64_FORMAT " on %s", pd->name, pd->partition.start, dest_slot->device);
+
+		if (!clear_boot_switch_partition(dest_slot->device, &pd->partition, &ierror)) {
+			g_propagate_error(error, ierror);
+			goto out;
+		}
+
+		if (!write_boot_switch_partition(image, dest_slot->device, &pd->partition, header_size, &ierror)) {
+			g_propagate_error(error, ierror);
+			goto out;
+		}
+	}
+
+	/* run slot post install hook if enabled */
+	if (hook_name && image->hooks.post_install) {
+		res = run_slot_hook_extra_env(hook_name, R_SLOT_HOOK_POST_INSTALL, image,
+				dest_slot, vars, &ierror);
+		if (!res) {
+			g_propagate_error(error, ierror);
+			goto out;
+		}
+	}
+
+out:
+	return res;
+}
+
 static gboolean img_to_raw_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	GError *ierror = NULL;
@@ -2028,8 +2511,8 @@ RaucUpdatePair updatepairs[] = {
 	{"*.squashfs-zst", "ubivol", img_to_ubivol_handler},
 #if ENABLE_EMMC_BOOT_SUPPORT == 1
 	{"*.img", "boot-emmc", img_to_boot_emmc_handler},
-	{"*", "boot-emmc", NULL},
 #endif
+	{"*", "boot-emmc", NULL},
 	{"*.vfat", "boot-mbr-switch", img_to_boot_mbr_switch_handler},
 	{"*.img", "boot-mbr-switch", img_to_boot_mbr_switch_handler},
 	{"*", "boot-mbr-switch", NULL},
@@ -2039,6 +2522,8 @@ RaucUpdatePair updatepairs[] = {
 	{"*.img", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
 #endif
 	{"*", "boot-gpt-switch", NULL},
+	{"*.img", "boot-raw-fallback", img_to_boot_raw_fallback_handler},
+	{"*", "boot-raw-fallback", NULL},
 	{"*.img.caibx", "*", img_to_raw_handler}, /* fallback */
 	{"*.img", "*", img_to_raw_handler}, /* fallback */
 	{0}
