@@ -4,8 +4,10 @@
 #include <glib.h>
 #include <glib/gstdio.h>
 #include <limits.h>
+#include <linux/fs.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/ioctl.h>
 #include <unistd.h>
 
 #include "utils.h"
@@ -74,6 +76,10 @@ gboolean write_file(const gchar *filename, GBytes *bytes, GError **error)
 	const gchar *contents;
 	gsize length;
 
+	g_return_val_if_fail(filename != NULL, FALSE);
+	g_return_val_if_fail(bytes != NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	contents = g_bytes_get_data(bytes, &length);
 
 	return g_file_set_contents(filename, contents, length, error);
@@ -128,8 +134,7 @@ gboolean rm_tree(const gchar *path, GError **error)
 	return TRUE;
 }
 
-
-gchar *resolve_path(const gchar *basefile, gchar *path)
+gchar *resolve_path(const gchar *basefile, const gchar *path)
 {
 	g_autofree gchar *cwd = NULL;
 	g_autofree gchar *dir = NULL;
@@ -152,10 +157,17 @@ gchar *resolve_path(const gchar *basefile, gchar *path)
 	return g_build_filename(cwd, dir, path, NULL);
 }
 
+gchar *resolve_path_take(const char *basefile, gchar *path)
+{
+	gchar *result = resolve_path(basefile, path);
+	g_free(path);
+	return result;
+}
+
 gboolean check_remaining_groups(GKeyFile *key_file, GError **error)
 {
 	gsize rem_num_groups;
-	gchar **rem_groups;
+	g_auto(GStrv) rem_groups = NULL;
 
 	rem_groups = g_key_file_get_groups(key_file, &rem_num_groups);
 	if (rem_num_groups != 0) {
@@ -170,7 +182,7 @@ gboolean check_remaining_groups(GKeyFile *key_file, GError **error)
 gboolean check_remaining_keys(GKeyFile *key_file, const gchar *groupname, GError **error)
 {
 	gsize rem_num_keys;
-	gchar **rem_keys;
+	g_auto(GStrv) rem_keys = NULL;
 
 	rem_keys = g_key_file_get_keys(key_file, groupname, &rem_num_keys, NULL);
 	if (rem_keys && rem_num_keys != 0) {
@@ -340,3 +352,142 @@ gchar *r_hex_encode(const guint8 *raw, size_t len)
 
 	return hex;
 }
+
+gboolean r_read_exact(const int fd, guint8 *data, size_t size, GError **error)
+{
+	size_t pos = 0;
+
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	while (pos < size) {
+		size_t remaining = size - pos;
+		ssize_t ret = TEMP_FAILURE_RETRY(read(fd, data+pos, remaining));
+		if (ret < 0) {
+			int err = errno;
+			g_set_error(error,
+					G_FILE_ERROR,
+					g_file_error_from_errno(err),
+					"Failed to read: %s", g_strerror(err));
+			return FALSE;
+		} else if (ret == 0) { /* end of file */
+			return FALSE;
+		} else if ((size_t)ret <= remaining) {
+			pos += ret;
+		} else if ((size_t)ret > remaining) {
+			g_assert_not_reached();
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
+gboolean r_write_exact(const int fd, const guint8 *data, size_t size, GError **error)
+{
+	size_t pos = 0;
+
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	while (pos < size) {
+		size_t remaining = size - pos;
+		ssize_t ret = TEMP_FAILURE_RETRY(write(fd, data+pos, remaining));
+		if (ret < 0) {
+			int err = errno;
+			g_set_error(error,
+					G_FILE_ERROR,
+					g_file_error_from_errno(err),
+					"Failed to write: %s", g_strerror(err));
+			return FALSE;
+		} else if ((size_t)ret <= remaining) {
+			pos += ret;
+		} else if ((size_t)ret > remaining) {
+			g_assert_not_reached();
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
+gboolean r_pread_exact(const int fd, guint8 *data, size_t size, off_t offset, GError **error)
+{
+	size_t pos = 0;
+
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	while (pos < size) {
+		size_t remaining = size - pos;
+		ssize_t ret = TEMP_FAILURE_RETRY(pread(fd, data+pos, remaining, offset+pos));
+		if (ret < 0) {
+			int err = errno;
+			g_set_error(error,
+					G_FILE_ERROR,
+					g_file_error_from_errno(err),
+					"Failed to read: %s", g_strerror(err));
+			return FALSE;
+		} else if (ret == 0) { /* end of file */
+			return FALSE;
+		} else if ((size_t)ret <= remaining) {
+			pos += ret;
+		} else if ((size_t)ret > remaining) {
+			g_assert_not_reached();
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
+gboolean r_pwrite_exact(const int fd, const guint8 *data, size_t size, off_t offset, GError **error)
+{
+	size_t pos = 0;
+
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	while (pos < size) {
+		size_t remaining = size - pos;
+		ssize_t ret = TEMP_FAILURE_RETRY(pwrite(fd, data+pos, remaining, offset+pos));
+		if (ret < 0) {
+			int err = errno;
+			g_set_error(error,
+					G_FILE_ERROR,
+					g_file_error_from_errno(err),
+					"Failed to write: %s", g_strerror(err));
+			return FALSE;
+		} else if ((size_t)ret <= remaining) {
+			pos += ret;
+		} else if ((size_t)ret > remaining) {
+			g_assert_not_reached();
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
+gboolean r_pwrite_lazy(const int fd, const guint8 *data, size_t size, off_t offset, GError **error)
+{
+	g_autofree guint8 *read_data = g_malloc(size);
+	GError *ierror = NULL;
+
+	if (!r_pread_exact(fd, read_data, size, offset, &ierror)) {
+		g_propagate_prefixed_error(error, ierror, "Failed to read existing data: ");
+		return FALSE;
+	}
+
+	if (memcmp(data, read_data, size) == 0) {
+		return TRUE;
+	}
+
+	return r_pwrite_exact(fd, data, size, offset, error);
+}
+
+guint get_sectorsize(gint fd)
+{
+	guint sector_size;
+
+	if (ioctl(fd, BLKSSZGET, &sector_size) != 0)
+		return 512;
+
+	return sector_size;
+}
