@@ -88,6 +88,7 @@ static void bundle_fixture_set_up_bundle_email(BundleFixture *fixture,
 	r_context_conf()->certpath = g_strdup("test/openssl-ca/dev/xku-emailProtection.cert.pem");
 	r_context_conf()->keypath = g_strdup("test/openssl-ca/dev/private/xku-emailProtection.pem");
 	/* cert is already checked once during signing */
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("smimesign");
 
 	prepare_bundle(fixture, user_data);
@@ -99,6 +100,7 @@ static void bundle_fixture_set_up_bundle_codesign(BundleFixture *fixture,
 	r_context_conf()->certpath = g_strdup("test/openssl-ca/dev/xku-codeSigning.cert.pem");
 	r_context_conf()->keypath = g_strdup("test/openssl-ca/dev/private/xku-codeSigning.pem");
 	/* cert is already checked once during signing */
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("codesign");
 
 	prepare_bundle(fixture, user_data);
@@ -130,9 +132,9 @@ static void test_check_empty_bundle(BundleFixture *fixture,
 	bundlename = write_random_file(fixture->tmpdir, "bundle.raucb", 0, 1234);
 	g_assert_nonnull(bundlename);
 
-	res = check_bundle(bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_false(res);
-	g_assert_error(ierror, G_IO_ERROR, G_IO_ERROR_PARTIAL_INPUT);
+	g_assert_error(ierror, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
 	g_assert_null(bundle);
 }
 
@@ -147,9 +149,9 @@ static void test_check_invalid_bundle(BundleFixture *fixture,
 	bundlename = write_random_file(fixture->tmpdir, "bundle.raucb", 1024, 1234);
 	g_assert_nonnull(bundlename);
 
-	res = check_bundle(bundlename, &bundle, CHECK_BUNDLE_NO_VERIFY, &ierror);
+	res = check_bundle(bundlename, &bundle, CHECK_BUNDLE_NO_VERIFY, NULL, &ierror);
 	g_assert_false(res);
-	g_assert_error(ierror, R_BUNDLE_ERROR, R_BUNDLE_ERROR_IDENTIFIER);
+	g_assert_error(ierror, R_BUNDLE_ERROR, R_BUNDLE_ERROR_SIGNATURE);
 	g_assert_null(bundle);
 }
 
@@ -160,7 +162,7 @@ static void bundle_test_create_check_error(BundleFixture *fixture,
 	g_autoptr(GError) ierror = NULL;
 	gboolean res = FALSE;
 
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_false(res);
 	g_assert_error(ierror, R_BUNDLE_ERROR, R_BUNDLE_ERROR_FORMAT);
 	g_assert_null(bundle);
@@ -178,7 +180,7 @@ static void bundle_test_create_extract(BundleFixture *fixture,
 	outputdir = g_build_filename(fixture->tmpdir, "output", NULL);
 	g_assert_nonnull(outputdir);
 
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
@@ -211,7 +213,7 @@ static void bundle_test_create_mount_extract(BundleFixture *fixture,
 	if (!test_running_as_root())
 		return;
 
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_NO_VERIFY, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_NO_VERIFY, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
@@ -236,7 +238,7 @@ static void bundle_test_extract_signature(BundleFixture *fixture,
 	outputsig = g_build_filename(fixture->tmpdir, "bundle.sig", NULL);
 	g_assert_nonnull(outputsig);
 
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
@@ -285,7 +287,7 @@ static void bundle_test_check_casync_old(BundleFixture *fixture, gconstpointer u
 	g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
 			"Verifying bundle signature*" );
 
-	res = check_bundle(bundlepath, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(bundlepath, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
@@ -314,7 +316,7 @@ static void bundle_test_check_casync_new(BundleFixture *fixture, gconstpointer u
 	g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
 			"Verifying bundle signature*" );
 
-	res = check_bundle(bundlepath, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(bundlepath, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
@@ -346,7 +348,7 @@ static void bundle_test_replace_signature(BundleFixture *fixture,
 	g_assert_nonnull(resignbundle);
 
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/rel-ca.pem");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
@@ -354,7 +356,7 @@ static void bundle_test_replace_signature(BundleFixture *fixture,
 
 	/* Verify input bundle with 'dev' keyring */
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/dev-only-ca.pem");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -370,13 +372,13 @@ static void bundle_test_replace_signature(BundleFixture *fixture,
 	g_clear_pointer(&bundle, free_bundle);
 
 	/* Verify resigned bundle with old 'dev' key */
-	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/rel-ca.pem");
-	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -391,7 +393,7 @@ static void bundle_test_replace_signature(BundleFixture *fixture,
 	g_clear_pointer(&bundle, free_bundle);
 
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/dev-only-ca.pem");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -400,21 +402,21 @@ static void bundle_test_replace_signature(BundleFixture *fixture,
 	g_assert_true(res);
 	g_clear_pointer(&bundle, free_bundle);
 
-	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 	g_clear_pointer(&bundle, free_bundle);
 
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/rel-ca.pem");
-	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_clear_pointer(&bundle, free_bundle);
 
 	/* Test without verify */
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/dev-only-ca.pem");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -432,14 +434,14 @@ static void bundle_test_replace_signature(BundleFixture *fixture,
 	g_clear_pointer(&bundle, free_bundle);
 	g_clear_pointer(&sigpath, g_free);
 
-	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_pointer(&bundle, free_bundle);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/rel-ca.pem");
-	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, &ierror);
+	res = check_bundle(replacebundle, &bundle, CHECK_BUNDLE_TRUST_ENV, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -463,7 +465,7 @@ static void bundle_test_resign(BundleFixture *fixture,
 	 * the context's 'pending' flag which would cause a re-initialization
 	 * of context and thus overwrite content of 'config' member. */
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/rel-ca.pem");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
@@ -472,7 +474,7 @@ static void bundle_test_resign(BundleFixture *fixture,
 
 	/* Verify input bundle with 'dev' keyring */
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/dev-only-ca.pem");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -493,7 +495,7 @@ static void bundle_test_resign(BundleFixture *fixture,
 	 * installing development bundles as well as moving to production
 	 * bundles. */
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/dev-only-ca.pem");
-	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
@@ -502,7 +504,7 @@ static void bundle_test_resign(BundleFixture *fixture,
 
 	/* Verify resigned bundle with rel keyring */
 	r_context()->config->keyring_path = g_strdup("test/openssl-ca/rel-ca.pem");
-	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(resignbundle, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 
@@ -517,7 +519,7 @@ static void bundle_test_wrong_capath(BundleFixture *fixture,
 	g_autoptr(GError) ierror = NULL;
 	r_context()->config->keyring_path = g_strdup("does/not/exist.pem");
 
-	g_assert_false(check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror));
+	g_assert_false(check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror));
 	g_assert_null(bundle);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_CA_LOAD);
 
@@ -540,7 +542,7 @@ static void bundle_test_verify_no_crl_warn(BundleFixture *fixture,
 			"Reading bundle*");
 	g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING,
 			"Detected CRL but CRL checking is disabled!");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
@@ -554,7 +556,7 @@ static void bundle_test_verify_revoked(BundleFixture *fixture,
 	g_autoptr(RaucBundle) bundle = NULL;
 	g_autoptr(GError) ierror = NULL;
 
-	g_assert_false(check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror));
+	g_assert_false(check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror));
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_assert_null(bundle);
 }
@@ -569,36 +571,41 @@ static void bundle_test_purpose_default(BundleFixture *fixture,
 	/* When the cert specifies no purpose, everything except 'codesign' is allowed */
 
 	g_message("testing default purpose with default cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = NULL;
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
 	g_message("testing purpose 'smimesign' with default cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("smimesign");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
 	g_message("testing purpose 'codesign' with default cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("codesign");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 
 	g_message("testing purpose 'any' with default cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("any");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = NULL;
 }
 
@@ -612,37 +619,42 @@ static void bundle_test_purpose_email(BundleFixture *fixture,
 	/* When the cert specifies the 'smimesign' usage, only default and that is allowed */
 
 	g_message("testing default purpose with 'smimesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = NULL;
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
 	g_message("testing purpose 'smimesign' with 'smimesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("smimesign");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
 	g_message("testing purpose 'codesign' with 'smimesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("codesign");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 	g_assert_null(bundle);
 
 	g_message("testing purpose 'any' with 'smimesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("any");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = NULL;
 }
 
@@ -656,37 +668,42 @@ static void bundle_test_purpose_codesign(BundleFixture *fixture,
 	/* When the cert specifies the 'codesign' usage, only that is allowed */
 
 	g_message("testing default purpose with 'codesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = NULL;
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 	g_assert_null(bundle);
 
 	g_message("testing purpose 'smimesign' with 'codesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("smimesign");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_error(ierror, R_SIGNATURE_ERROR, R_SIGNATURE_ERROR_INVALID);
 	g_clear_error(&ierror);
 	g_assert_false(res);
 	g_assert_null(bundle);
 
 	g_message("testing purpose 'codesign' with 'codesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("codesign");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
 	g_message("testing purpose 'any' with 'codesign' cert");
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = g_strdup("any");
-	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, &ierror);
+	res = check_bundle(fixture->bundlename, &bundle, CHECK_BUNDLE_DEFAULT, NULL, &ierror);
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(bundle);
 	g_clear_pointer(&bundle, free_bundle);
 
+	g_free(r_context()->config->keyring_check_purpose);
 	r_context()->config->keyring_check_purpose = NULL;
 }
 
