@@ -18,7 +18,7 @@ typedef struct {
 guint8* random_bytes(gsize size, guint32 seed)
 {
 	guint8 *bytes = g_malloc0(size);
-	GRand *rand = g_rand_new_with_seed(seed);
+	g_autoptr(GRand) rand = g_rand_new_with_seed(seed);
 	for (gsize i = 0; i < size; i++) {
 		bytes[i] = g_rand_int(rand) & 0xFF;
 	}
@@ -29,7 +29,7 @@ gchar* write_random_file(const gchar *tmpdir, const gchar *filename,
 		gsize size, const guint32 seed)
 {
 	gchar *pathname;
-	guint8 *content;
+	g_autofree guint8 *content = NULL;
 
 	pathname = g_build_filename(tmpdir, filename, NULL);
 	g_assert_nonnull(pathname);
@@ -40,7 +40,6 @@ gchar* write_random_file(const gchar *tmpdir, const gchar *filename,
 		return NULL;
 	}
 
-	g_free(content);
 	return pathname;
 }
 
@@ -251,10 +250,6 @@ gboolean test_do_chmod(const gchar *path)
 	gboolean res = FALSE;
 	GPtrArray *args = g_ptr_array_new_full(10, g_free);
 
-	if (getuid() != 0) {
-		g_ptr_array_add(args, g_strdup("sudo"));
-		g_ptr_array_add(args, g_strdup("--non-interactive"));
-	}
 	g_ptr_array_add(args, g_strdup("chmod"));
 	g_ptr_array_add(args, g_strdup("777"));
 	g_ptr_array_add(args, g_strdup(path));
@@ -359,3 +354,40 @@ gboolean test_running_as_root(void)
 	g_test_skip("not running as root");
 	return FALSE;
 }
+
+gsize get_file_size(gchar* filename, GError **error)
+{
+	GError *ierror = NULL;
+	GFile *file = NULL;
+	GFileInputStream *filestream = NULL;
+	gsize size = 0;
+	gboolean res = FALSE;
+
+	file = g_file_new_for_path(filename);
+	filestream = g_file_read(file, NULL, &ierror);
+	if (filestream == NULL) {
+		g_propagate_prefixed_error(
+				error,
+				ierror,
+				"failed to open bundle for reading: ");
+		goto out;
+	}
+
+	res = g_seekable_seek(G_SEEKABLE(filestream),
+			0, G_SEEK_END, NULL, &ierror);
+	if (!res) {
+		g_propagate_prefixed_error(
+				error,
+				ierror,
+				"failed to seek to end of bundle: ");
+		goto out;
+	}
+
+	size = g_seekable_tell(G_SEEKABLE(filestream));
+
+out:
+	g_clear_object(&filestream);
+	g_clear_object(&file);
+
+	return size;
+}
