@@ -68,6 +68,7 @@ path=/etc/rauc/keyring/\n\
 [casync]\n\
 storepath=/var/lib/default.castr/\n\
 tmppath=/tmp/\n\
+install-args=--verbose\n\
 \n\
 [slot.rescue.0]\n\
 description=Rescue partition\n\
@@ -120,6 +121,10 @@ install-same=false\n";
 	g_assert_cmpstr(config->statusfile_path, ==, "/mnt/persistent-rw-fs/system.raucs");
 	g_assert_cmpint(config->max_bundle_download_size, ==, 42);
 	g_assert_cmphex(config->bundle_formats_mask, ==, 0x2);
+	g_assert_cmpstr(config->store_path, ==, "/var/lib/default.castr/");
+	g_assert_cmpstr(config->tmp_path, ==, "/tmp/");
+	g_assert_cmpstr(config->casync_install_args, ==, "--verbose");
+	g_assert_false(config->use_desync);
 
 	g_assert_nonnull(config->slots);
 	slotlist = g_hash_table_get_keys(config->slots);
@@ -1009,6 +1014,7 @@ compatible=FooCorp Super BarBazzer\n\
 bootloader=barebox\n\
 [keyring]\n\
 path=/dev/null\n\
+allow-partial-chain=true\n\
 check-crl=true\n\
 check-purpose=codesign\n";
 
@@ -1019,6 +1025,7 @@ check-purpose=codesign\n";
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(config);
+	g_assert_false(config->keyring_allow_partial_chain);
 	g_assert_false(config->keyring_check_crl);
 	g_assert_cmpstr(config->keyring_check_purpose, ==, NULL);
 
@@ -1031,6 +1038,7 @@ check-purpose=codesign\n";
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(config);
+	g_assert_true(config->keyring_allow_partial_chain);
 	g_assert_true(config->keyring_check_crl);
 	g_assert_cmpstr(config->keyring_check_purpose, ==, "codesign");
 
@@ -1063,7 +1071,7 @@ bundle-formats=-plain\n";
 [system]\n\
 compatible=FooCorp Super BarBazzer\n\
 bootloader=barebox\n\
-bundle-formats=-plain -verity\n";
+bundle-formats=-plain -verity -crypt\n";
 
 	pathname = write_tmp_file(fixture->tmpdir, "default.conf", default_cfg_file, NULL);
 	g_assert_nonnull(pathname);
@@ -1072,7 +1080,7 @@ bundle-formats=-plain -verity\n";
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(config);
-	g_assert_cmphex(config->bundle_formats_mask, ==, 0x3);
+	g_assert_cmphex(config->bundle_formats_mask, ==, 0x7);
 
 	free_config(config);
 
@@ -1094,7 +1102,7 @@ bundle-formats=-plain -verity\n";
 	g_assert_no_error(ierror);
 	g_assert_true(res);
 	g_assert_nonnull(config);
-	g_assert_cmphex(config->bundle_formats_mask, ==, 0x2);
+	g_assert_cmphex(config->bundle_formats_mask, ==, 0x6);
 
 	free_config(config);
 
@@ -1103,7 +1111,7 @@ bundle-formats=-plain -verity\n";
 
 	res = load_config(pathname, &config, &ierror);
 	g_assert_error(ierror, R_CONFIG_ERROR, R_CONFIG_ERROR_INVALID_FORMAT);
-	g_assert_cmpstr(ierror->message, ==, "Invalid bundle format configuration '-plain -verity', no remaining formats");
+	g_assert_cmpstr(ierror->message, ==, "Invalid bundle format configuration '-plain -verity -crypt', no remaining formats");
 	g_assert_false(res);
 	g_assert_null(config);
 	g_clear_error(&ierror);
@@ -1162,6 +1170,40 @@ static void config_file_test_parse_bundle_formats(void)
 	g_clear_error(&ierror);
 }
 
+static void config_file_use_desync_set_to_true(ConfigFileFixture *fixture,
+		gconstpointer user_data)
+{
+	RaucConfig *config;
+	g_autoptr(GError) ierror = NULL;
+	gboolean res;
+	g_autofree gchar* pathname = NULL;
+
+	const gchar *cfg_file = "\
+[system]\n\
+compatible=FooCorp Super BarBazzer\n\
+bootloader=barebox\n\
+\n\
+[casync]\n\
+storepath=/var/lib/default.castr/\n\
+tmppath=/tmp/\n\
+install-args=--seed /my/path/additional_seed.caibx\n\
+use-desync=true";
+
+	pathname = write_tmp_file(fixture->tmpdir, "simple_desync.conf", cfg_file, NULL);
+	g_assert_nonnull(pathname);
+
+	res = load_config(pathname, &config, &ierror);
+	g_assert_no_error(ierror);
+	g_assert_true(res);
+	g_assert_nonnull(config);
+	g_assert_cmpstr(config->store_path, ==, "/var/lib/default.castr/");
+	g_assert_cmpstr(config->tmp_path, ==, "/tmp/");
+	g_assert_cmpstr(config->casync_install_args, ==, "--seed /my/path/additional_seed.caibx");
+	g_assert_true(config->use_desync);
+
+	free_config(config);
+}
+
 int main(int argc, char *argv[])
 {
 	setlocale(LC_ALL, "C");
@@ -1258,6 +1300,9 @@ int main(int argc, char *argv[])
 			config_file_fixture_set_up, config_file_bundle_formats,
 			config_file_fixture_tear_down);
 	g_test_add_func("/config-file/parse-bundle-formats", config_file_test_parse_bundle_formats);
+	g_test_add("/config-file/use-desync", ConfigFileFixture, NULL,
+			config_file_fixture_set_up, config_file_use_desync_set_to_true,
+			config_file_fixture_tear_down);
 
 	return g_test_run();
 }
