@@ -9,10 +9,10 @@
 void fixture_helper_fixture_set_up_system_user(gchar *tmpdir,
 		const gchar *configname)
 {
-	gchar *configpath;
-	gchar *certpath;
-	gchar *keypath;
-	gchar *capath;
+	g_autofree gchar *configpath = NULL;
+	g_autofree gchar *certpath = NULL;
+	g_autofree gchar *keypath = NULL;
+	g_autofree gchar *capath = NULL;
 
 	g_assert_nonnull(tmpdir);
 	g_print("bundle tmpdir: %s\n", tmpdir);
@@ -78,11 +78,6 @@ void fixture_helper_fixture_set_up_system_user(gchar *tmpdir,
 
 	/* Set dummy bootname provider */
 	r_context_conf()->bootslot = g_strdup("system0");
-
-	g_free(configpath);
-	g_free(certpath);
-	g_free(keypath);
-	g_free(capath);
 }
 
 void fixture_helper_set_up_system(gchar *tmpdir,
@@ -121,13 +116,14 @@ void fixture_helper_set_up_system(gchar *tmpdir,
 
 void fixture_helper_set_up_bundle(gchar *tmpdir,
 		const gchar* manifest_content,
-		gboolean handler,
-		gboolean hook)
+		const ManifestTestOptions *options)
 {
-	gchar *contentdir;
-	gchar *bundlepath;
-	gchar *mountdir;
-	gchar *testfilepath;
+	g_autofree gchar *contentdir = NULL;
+	g_autofree gchar *bundlepath = NULL;
+	g_autofree gchar *mountdir = NULL;
+	g_autofree gchar *testfilepath = NULL;
+	g_autoptr(GError) error = NULL;
+	gboolean res = FALSE;
 
 	/* needs to run as root */
 	if (!test_running_as_root())
@@ -153,7 +149,7 @@ void fixture_helper_set_up_bundle(gchar *tmpdir,
 	if (manifest_content) {
 		g_assert_true(write_tmp_file(tmpdir, "content/manifest.raucm", manifest_content, NULL));
 	} else {
-		g_assert(test_prepare_manifest_file(tmpdir, "content/manifest.raucm", FALSE, hook) == 0);
+		g_assert(test_prepare_manifest_file(tmpdir, "content/manifest.raucm", options) == 0);
 	}
 
 	/* Make images user-writable */
@@ -169,25 +165,19 @@ void fixture_helper_set_up_bundle(gchar *tmpdir,
 	g_assert(test_rmdir(tmpdir, "mnt") == 0);
 
 	/* Copy custom handler */
-	if (handler) {
+	if (options->custom_handler) {
 		g_assert_true(test_copy_file("test/install-content/custom_handler.sh", NULL,
 				tmpdir, "content/custom_handler.sh"));
 	}
 
 	/* Copy hook */
-	if (hook) {
+	if (options->hooks) {
 		g_assert_true(test_copy_file("test/install-content/hook.sh", NULL,
 				tmpdir, "content/hook.sh"));
 	}
 
-	/* Update checksums in manifest */
-	g_assert_true(update_manifest(contentdir, FALSE, NULL));
-
 	/* Create bundle */
-	g_assert_true(create_bundle(bundlepath, contentdir, NULL));
-
-	g_free(bundlepath);
-	g_free(contentdir);
-	g_free(mountdir);
-	g_free(testfilepath);
+	res = create_bundle(bundlepath, contentdir, &error);
+	g_assert_no_error(error);
+	g_assert_true(res);
 }
