@@ -168,6 +168,7 @@ echo "Release Signing Key"
 cd $BASE/rel
 openssl req -newkey rsa -keyout private/release-1.pem -out release-1.csr.pem -subj "/O=$ORG/CN=$ORG Release-1"
 openssl ca -batch -extensions v3_leaf -in release-1.csr.pem -out release-1.cert.pem
+openssl rsa -aes256 -in private/release-1.pem -out private/release-1-encrypted.pem -passout pass:1111
 openssl req -newkey rsa -keyout private/release-2018.pem -out release-2018.csr.pem -subj "/O=$ORG/CN=$ORG Release-2018"
 openssl ca -batch -extensions v3_leaf -in release-2018.csr.pem -out release-2018.cert.pem -startdate 20180101000000Z -enddate 20190701000000Z
 
@@ -216,6 +217,8 @@ cat root/ca.cert.pem root/crl.pem rel/ca.cert.pem rel/crl.pem dev/ca.cert.pem de
 cat root/ca.cert.pem root/crl.pem dev/ca.cert.pem dev/crl.pem > dev-only-ca.pem
 cat root/ca.cert.pem root/crl.pem rel/ca.cert.pem rel/crl.pem > rel-ca.pem
 cat root/ca.cert.pem root/crl.pem web/ca.cert.pem web/crl.pem > web-ca.pem
+cat dev/ca.cert.pem > dev-partial-ca.pem
+cat rel/ca.cert.pem > rel-partial-ca.pem
 
 echo "Build Directory Test Keys"
 cd $BASE/dir
