@@ -9,22 +9,24 @@
 #include "context.h"
 #include "mount.h"
 #include "utils.h"
+#include "stats.h"
 
 typedef struct {
 	gchar *tmpdir;
 } UpdateHandlerFixture;
 
 typedef enum {
-	TEST_UPDATE_HANDLER_DEFAULT       = 0,
-	TEST_UPDATE_HANDLER_EXPECT_FAIL   = BIT(0),
-	TEST_UPDATE_HANDLER_NO_IMAGE_FILE = BIT(1),
-	TEST_UPDATE_HANDLER_NO_TARGET_DEV = BIT(2),
-	TEST_UPDATE_HANDLER_HOOKS         = BIT(3),
-	TEST_UPDATE_HANDLER_PRE_HOOK      = BIT(4),
-	TEST_UPDATE_HANDLER_POST_HOOK     = BIT(5),
-	TEST_UPDATE_HANDLER_INSTALL_HOOK  = BIT(6),
-	TEST_UPDATE_HANDLER_NO_HOOK_FILE  = BIT(7),
-	TEST_UPDATE_HANDLER_HOOK_FAIL     = BIT(8),
+	TEST_UPDATE_HANDLER_DEFAULT                                       = 0,
+	TEST_UPDATE_HANDLER_EXPECT_FAIL                                   = BIT(0),
+	TEST_UPDATE_HANDLER_NO_IMAGE_FILE                                 = BIT(1),
+	TEST_UPDATE_HANDLER_NO_TARGET_DEV                                 = BIT(2),
+	TEST_UPDATE_HANDLER_HOOKS                                         = BIT(3),
+	TEST_UPDATE_HANDLER_PRE_HOOK                                      = BIT(4),
+	TEST_UPDATE_HANDLER_POST_HOOK                                     = BIT(5),
+	TEST_UPDATE_HANDLER_INSTALL_HOOK                                  = BIT(6),
+	TEST_UPDATE_HANDLER_NO_HOOK_FILE                                  = BIT(7),
+	TEST_UPDATE_HANDLER_HOOK_FAIL                                     = BIT(8),
+	TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX                           = BIT(9),
 } TestUpdateHandlerParams;
 
 typedef struct {
@@ -129,46 +131,12 @@ static void update_handler_fixture_tear_down(UpdateHandlerFixture *fixture,
 	if (!(test_pair->params & TEST_UPDATE_HANDLER_NO_TARGET_DEV)) {
 		g_assert(test_remove(fixture->tmpdir, "rootfs-0") == 0);
 	}
+	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+		test_rm_tree(fixture->tmpdir, "rootfs-0-datadir");
+	}
 	g_assert(test_rmdir(fixture->tmpdir, "") == 0);
 }
 
-static gsize get_file_size(gchar* filename, GError **error)
-{
-	GError *ierror = NULL;
-	GFile *file = NULL;
-	GFileInputStream *filestream = NULL;
-	gsize size = 0;
-	gboolean res = FALSE;
-
-	file = g_file_new_for_path(filename);
-	filestream = g_file_read(file, NULL, &ierror);
-	if (filestream == NULL) {
-		g_propagate_prefixed_error(
-				error,
-				ierror,
-				"failed to open bundle for reading: ");
-		goto out;
-	}
-
-	res = g_seekable_seek(G_SEEKABLE(filestream),
-			0, G_SEEK_END, NULL, &ierror);
-	if (!res) {
-		g_propagate_prefixed_error(
-				error,
-				ierror,
-				"failed to seek to end of bundle: ");
-		goto out;
-	}
-
-	size = g_seekable_tell(G_SEEKABLE(filestream));
-
-out:
-	g_clear_object(&filestream);
-	g_clear_object(&file);
-
-	return size;
-}
-
 static gboolean tar_image(const gchar *dest, const gchar *dir, GError **error)
 {
 	GSubprocess *sproc = NULL;
@@ -209,6 +177,46 @@ out:
 	return res;
 }
 
+static gboolean casync_blob_image(const gchar *idxpath, const gchar *contentpath, const gchar *store, GError **error)
+{
+	GSubprocess *sproc = NULL;
+	GError *ierror = NULL;
+	gboolean res = FALSE;
+	GPtrArray *args = g_ptr_array_new_full(5, g_free);
+
+	g_ptr_array_add(args, g_strdup("casync"));
+	g_ptr_array_add(args, g_strdup("make"));
+	g_ptr_array_add(args, g_strdup("--with=unix"));
+	g_ptr_array_add(args, g_strdup(idxpath));
+	g_ptr_array_add(args, g_strdup(contentpath));
+	g_ptr_array_add(args, g_strdup("--store"));
+	g_ptr_array_add(args, g_strdup(store));
+	g_ptr_array_add(args, NULL);
+
+	sproc = r_subprocess_newv(args, G_SUBPROCESS_FLAGS_STDOUT_SILENCE, &ierror);
+	if (sproc == NULL) {
+		g_propagate_prefixed_error(
+				error,
+				ierror,
+				"failed to start casync: ");
+		goto out;
+	}
+
+	res = g_subprocess_wait_check(sproc, NULL, &ierror);
+	if (!res) {
+		g_propagate_prefixed_error(
+				error,
+				ierror,
+				"failed to run casync: ");
+		goto out;
+	}
+
+out:
+	g_ptr_array_unref(args);
+	g_clear_object(&sproc);
+	return res;
+}
+
 /**
  * Create dummy archive.
  *
@@ -246,6 +254,51 @@ out:
 	return res;
 }
 
+static gchar* test_prepare_dummy_caibx(const gchar *path, const gchar *archname)
+{
+	g_autoptr(GError) ierror = NULL;
+	g_autofree gchar *idxpath = NULL, *storepath = NULL;
+	g_autofree gchar* pathname = NULL;
+
+	idxpath = g_build_filename(path, archname, NULL);
+	storepath = g_build_filename(path, "out.castr", NULL);
+	pathname = write_random_file(path, "tmp.img", IMAGE_SIZE, 0xe92001ca);
+	g_assert_nonnull(pathname);
+
+	if (!casync_blob_image(idxpath, pathname, storepath, &ierror)) {
+		g_warning("%s", ierror->message);
+		return NULL;
+	}
+
+	g_assert(test_remove(path, "tmp.img") == 0);
+	return g_steal_pointer(&storepath);
+}
+
+static gchar* test_prepare_dummy_caidx(const gchar *path, const gchar *archname)
+{
+	g_autoptr(GError) ierror = NULL;
+	g_autofree gchar *idxpath = NULL, *storepath = NULL;
+	g_autofree gchar* contentpath = NULL;
+
+	idxpath = g_build_filename(path, archname, NULL);
+	storepath = g_build_filename(path, "out.castr", NULL);
+
+	contentpath = g_build_filename(path, "content", NULL);
+
+	g_assert(g_mkdir(contentpath, 0777) == 0);
+	g_assert(test_prepare_dummy_file(contentpath, "testfile.txt",
+			FILE_SIZE, "/dev/zero") == 0);
+
+	if (!casync_blob_image(idxpath, contentpath, storepath, &ierror)) {
+		g_warning("%s", ierror->message);
+		return NULL;
+	}
+
+	g_assert(test_remove(contentpath, "testfile.txt") == 0);
+	g_assert(g_rmdir(contentpath) == 0);
+	return g_steal_pointer(&storepath);
+}
+
 static void test_update_handler(UpdateHandlerFixture *fixture,
 		gconstpointer user_data)
 {
@@ -292,6 +345,14 @@ static void test_update_handler(UpdateHandlerFixture *fixture,
 	}
 	imagepath = g_build_filename(fixture->tmpdir, imagename, NULL);
 
+	/* skip casync checks if casync is not available */
+	if ((g_strcmp0(test_pair->imagetype, "img.caibx") == 0) || (g_strcmp0(test_pair->imagetype, "caidx") == 0)) {
+		if (!g_getenv("RAUC_TEST_CASYNC")) {
+			g_test_skip("RAUC_TEST_CASYNC undefined");
+			return;
+		}
+	}
+
 	/* create source image */
 	image = g_new0(RaucImage, 1);
 	image->slotclass = g_strdup("rootfs");
@@ -311,20 +372,35 @@ static void test_update_handler(UpdateHandlerFixture *fixture,
 			test_do_chmod(hookpath);
 		}
 	}
+	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+		image->adaptive = g_strsplit("block-hash-index", " ", 0);
+	}
 
 	if (test_pair->params & TEST_UPDATE_HANDLER_NO_IMAGE_FILE) {
 		goto no_image;
 	}
 
 	if (g_strcmp0(test_pair->imagetype, "img") == 0) {
-		g_assert(test_prepare_dummy_file(fixture->tmpdir, "image.img",
-				IMAGE_SIZE, "/dev/zero") == 0);
+		g_autofree gchar* pathname = write_random_file(fixture->tmpdir, "image.img", IMAGE_SIZE, 0x2abff992);
+		g_assert_nonnull(pathname);
 	} else if (g_strcmp0(test_pair->imagetype, "ext4") == 0) {
 		g_assert(test_prepare_dummy_file(fixture->tmpdir, "image.ext4",
 				IMAGE_SIZE, "/dev/zero") == 0);
 		g_assert(test_make_filesystem(fixture->tmpdir, "image.ext4"));
 	} else if (g_strcmp0(test_pair->imagetype, "tar.bz2") == 0) {
 		g_assert_true(test_prepare_dummy_archive(fixture->tmpdir, "image.tar.bz2", "testfile.txt"));
+	} else if (g_strcmp0(test_pair->imagetype, "img.caibx") == 0) {
+		gchar* storepath = test_prepare_dummy_caibx(fixture->tmpdir, "image.img.caibx");
+		g_assert_nonnull(storepath);
+
+		r_context_conf()->install_info->mounted_bundle = g_new0(RaucBundle, 1);
+		r_context_conf()->install_info->mounted_bundle->storepath = storepath;
+	} else if (g_strcmp0(test_pair->imagetype, "caidx") == 0) {
+		gchar* storepath = test_prepare_dummy_caidx(fixture->tmpdir, "image.caidx");
+		g_assert_nonnull(storepath);
+
+		r_context_conf()->install_info->mounted_bundle = g_new0(RaucBundle, 1);
+		r_context_conf()->install_info->mounted_bundle->storepath = storepath;
 	} else {
 		g_assert_not_reached();
 	}
@@ -337,6 +413,9 @@ no_image:
 	targetslot->device = g_strdup(slotpath);
 	targetslot->type = g_strdup(test_pair->slottype);
 	targetslot->state = ST_INACTIVE;
+	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+		targetslot->data_directory = g_build_filename(fixture->tmpdir, "rootfs-0-datadir", NULL);
+	}
 
 	/* Set mount path to current temp dir */
 	mountprefix = g_build_filename(fixture->tmpdir, "testmount", NULL);
@@ -351,7 +430,9 @@ no_image:
 	g_assert_nonnull(handler);
 
 	/* Run to perform an update */
+	r_test_stats_start();
 	res = handler(image, targetslot, hookpath, &ierror);
+	r_test_stats_stop();
 
 	if (test_pair->params & TEST_UPDATE_HANDLER_EXPECT_FAIL) {
 		g_assert_error(ierror, test_pair->err_domain, test_pair->err_code);
@@ -374,7 +455,7 @@ no_image:
 		g_assert_cmpint(get_file_size(imagepath, NULL), ==, IMAGE_SIZE);
 		g_assert(test_mount(slotpath, mountprefix));
 		g_assert(r_umount(slotpath, NULL));
-	} else if (g_strcmp0(test_pair->imagetype, "tar.bz2") == 0) {
+	} else if ((g_strcmp0(test_pair->imagetype, "tar.bz2") == 0) || ((g_strcmp0(test_pair->imagetype, "caidx") == 0))) {
 		gchar *testpath = g_build_filename(mountprefix, "testfile.txt", NULL);
 		g_assert(test_mount(slotpath, mountprefix));
 		g_assert_true(g_file_test(testpath, G_FILE_TEST_IS_REGULAR));
@@ -382,6 +463,74 @@ no_image:
 		g_free(testpath);
 	}
 
+	/* check statistics */
+	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+		RaucStats *stats;
+		guint64 count_zero = 0;
+		guint64 sum_zero = 0;
+		guint64 count_target_written = 0;
+		guint64 sum_target_written = 0;
+		guint64 count_target = 0;
+		guint64 sum_target = 0;
+		guint64 count_source = 0;
+		guint64 sum_source = 0;
+
+		stats = r_test_stats_next();
+		g_assert_nonnull(stats);
+		g_assert_cmpstr(stats->label, ==, "zero chunk");
+		count_zero = stats->count;
+		sum_zero = stats->sum;
+
+		stats = r_test_stats_next();
+		g_assert_nonnull(stats);
+		g_assert_cmpstr(stats->label, ==, "target_slot_written (reusing source_image)");;
+		count_target_written = stats->count;
+		sum_target_written = stats->sum;
+
+		stats = r_test_stats_next();
+		g_assert_nonnull(stats);
+		g_assert_cmpstr(stats->label, ==, "target_slot");
+		count_target = stats->count;
+		sum_target = stats->sum;
+
+		stats = r_test_stats_next();
+		g_assert_nonnull(stats);
+		g_assert_cmpstr(stats->label, ==, "source_image");
+		count_source = stats->count;
+		sum_source = stats->sum;
+
+		/* all non-zero chunks must result in a lookup in target_slot_written */
+		g_assert_cmpint(count_zero + count_target_written, ==, IMAGE_SIZE/4096);
+
+		/* sum of all found chunks must equal total number of chunks */
+		g_assert_cmpint(sum_zero + sum_target_written + sum_target + sum_source, ==, IMAGE_SIZE/4096);
+
+		if (g_strcmp0(test_pair->imagetype, "img") == 0) {
+			/* for random data it is *very* unlikely:
+			 * - to find zero chunks
+			 * - to find reusable chunks */
+			g_assert_cmpint(sum_zero, ==, 0);
+			g_assert_cmpint(sum_target_written, ==, 0);
+			g_assert_cmpint(sum_target, ==, 0);
+			g_assert_cmpint(sum_source, ==, IMAGE_SIZE/4096);
+		} else if (g_strcmp0(test_pair->imagetype, "ext4") == 0) {
+			/* for a generated ext4, actual values seem to depend on
+			 * used mkfs configuration. We use minimal values here
+			 * that have proven to be valid on all test systems.
+			 */
+			g_assert_cmpint(sum_zero, >=, IMAGE_SIZE/4096 - 29);
+			g_assert_cmpint(sum_target_written, >=, 0);
+			g_assert_cmpint(sum_target, ==, 0);
+			g_assert_cmpint(sum_source, <=, 29);
+		}
+
+		/* Number of total lookups must not increase in lookup order */
+		g_assert_cmpint(count_target_written, <=, IMAGE_SIZE/4096);
+		g_assert_cmpint(count_target, <=, count_target_written);
+		g_assert_cmpint(count_source, <=, count_target);
+	}
+	g_assert_null(r_test_stats_next());
+
 out:
 	/* clean up source image if it was generated */
 	if (!(test_pair->params & TEST_UPDATE_HANDLER_NO_IMAGE_FILE)) {
@@ -391,6 +540,14 @@ out:
 			g_assert(g_remove(imagepath) == 0);
 		} else if (g_strcmp0(test_pair->imagetype, "tar.bz2") == 0) {
 			g_assert(test_remove(fixture->tmpdir, "image.tar.bz2") == 0);
+		} else if (g_strcmp0(test_pair->imagetype, "img.caibx") == 0) {
+			g_assert(test_remove(fixture->tmpdir, "image.img.caibx") == 0);
+			g_assert_true(rm_tree(r_context()->install_info->mounted_bundle->storepath, NULL));
+			free_bundle(r_context_conf()->install_info->mounted_bundle);
+		} else if (g_strcmp0(test_pair->imagetype, "caidx") == 0) {
+			g_assert(test_remove(fixture->tmpdir, "image.caidx") == 0);
+			g_assert_true(rm_tree(r_context()->install_info->mounted_bundle->storepath, NULL));
+			free_bundle(r_context_conf()->install_info->mounted_bundle);
 		}
 	}
 
@@ -489,6 +646,18 @@ int main(int argc, char *argv[])
 		/* nor tests */
 		{"nor", "img", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
 
+		/* adaptive tests */
+		{"raw", "img", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX, 0, 0},
+		{"ext4", "img", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX, 0, 0},
+		{"raw", "ext4", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX, 0, 0},
+
+		/* casync blob index tests */
+		{"ext4", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"raw", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		/* casync directory index tests */
+		{"ext4", "caidx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"vfat", "caidx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+
 		{0}
 	};
 	setlocale(LC_ALL, "C");
@@ -856,5 +1025,54 @@ int main(int argc, char *argv[])
 			test_update_handler,
 			update_handler_fixture_tear_down);
 
+	/* adaptive tests */
+	g_test_add("/update_handler/block_hash_index/img_to_raw",
+			UpdateHandlerFixture,
+			&testpair_matrix[56],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/block_hash_index/img_to_ext4",
+			UpdateHandlerFixture,
+			&testpair_matrix[57],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/block_hash_index/ext4_to_raw",
+			UpdateHandlerFixture,
+			&testpair_matrix[58],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	/* casync blox index tests */
+	g_test_add("/update_handler/casync/img.caibx_to_ext4",
+			UpdateHandlerFixture,
+			&testpair_matrix[59],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/casync/img.caibx_to_raw",
+			UpdateHandlerFixture,
+			&testpair_matrix[60],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	/* casync directory index tests */
+	g_test_add("/update_handler/casync/caidx_to_ext4",
+			UpdateHandlerFixture,
+			&testpair_matrix[61],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	/* FIXME: only works with "--without=sec-time" set for "casync make"*/
+	/*
+	   g_test_add("/update_handler/casync/caidx_to_vfat",
+	                UpdateHandlerFixture,
+	                &testpair_matrix[62],
+	                update_handler_fixture_set_up,
+	                test_update_handler,
+	                update_handler_fixture_tear_down);
+	 */
+
 	return g_test_run();
 }
